# -*- coding: utf-8 -*-
'''
A state module designed to enforce load-balancing configurations for F5 Big-IP entities.
    :maturity:      develop
    :platform:      f5_bigip_11.6
'''

# Import Python libs
from __future__ import absolute_import

import json

# Import 3rd-party libs
import salt.ext.six as six


#set up virtual function
def __virtual__():
    '''
    Only load if the bigip exec module is available in __salt__
    '''
    return 'bigip' if 'bigip.list_transaction' in __salt__ else False


def _load_result(response, ret):
    '''
    format the results of listing functions
    '''

    #were we able to connect?
    if response['code'] is None:
        ret['comment'] = response['content']
    #forbidden?
    elif response['code'] == 401:
        ret['comment'] = '401 Forbidden: Authentication required!'
    #Not found?
    elif response['code'] == 404:
        ret['comment'] = response['content']['message']
    #200?
    elif response['code'] == 200:
        ret['result'] = True
        ret['comment'] = 'Listing Current Configuration Only.  ' \
                         'Not action or changes occurred during the execution of this state.'
        ret['changes'] = response['content']
    #something bad
    else:
        ret['comment'] = response['content']['message']

    return ret


def _strip_key(dictionary, keyword):
    '''
    look for a certain key within a dictionary and nullify ti's contents, check within nested
    dictionaries and lists as well.  Certain attributes such as "generation" will change even
    when there were no changes made to the entity.
    '''

    for key, value in six.iteritems(dictionary):
        if key == keyword:
            dictionary[key] = None
        elif isinstance(value, dict):
            _strip_key(value, keyword)
        elif isinstance(value, list):
            for item in value:
                if isinstance(item, dict):
                    _strip_key(item, keyword)

    return dictionary


def _check_for_changes(entity_type, ret, existing, modified):
    '''
    take an existing entity and a modified entity and check for changes.
    '''

    ret['result'] = True

    #were there any changes? generation always changes, remove it.

    if isinstance(existing, dict) and isinstance(modified, dict):
        if 'generation' in modified['content'].keys():
            del modified['content']['generation']

        if 'generation' in existing['content'].keys():
            del existing['content']['generation']

        if cmp(modified['content'], existing['content']) == 0:
            ret['comment'] = '{entity_type} is currently enforced to the desired state.  No changes made.'.format(entity_type=entity_type)
        else:
            ret['comment'] = '{entity_type} was enforced to the desired state.  Note: Only parameters specified ' \
                             'were enforced. See changes for details.'.format(entity_type=entity_type)
            ret['changes']['old'] = existing['content']
            ret['changes']['new'] = modified['content']

    else:
        if cmp(modified, existing) == 0:
            ret['comment'] = '{entity_type} is currently enforced to the desired state.  No changes made.'.format(entity_type=entity_type)
        else:
            ret['comment'] = '{entity_type} was enforced to the desired state.  Note: Only parameters specified ' \
                             'were enforced. See changes for details.'.format(entity_type=entity_type)
            ret['changes']['old'] = existing
            ret['changes']['new'] = modified

    return ret


def _test_output(ret, action, params):
    '''
    For testing just output what the state will attempt to do without actually doing it.
    '''

    if action == 'list':
        ret['comment'] += 'The list action will just list an entity and will make no changes.\n'
    elif action == 'create' or action == 'add':
        ret['comment'] += 'The create action will attempt to create an entity if it does not already exist.\n'
    elif action == 'delete':
        ret['comment'] += 'The delete action will attempt to delete an existing entity if it exists.\n'
    elif action == 'manage':
        ret['comment'] += 'The manage action will create a new entity if it does not exist.  If it does exist, it will be enforced' \
                          'to the desired state.\n'
    elif action == 'modify':
        ret['comment'] += 'The modify action will attempt to modify an existing entity only if it exists.\n'

    ret['comment'] += 'An iControl REST Request will be made using the parameters:\n'
    ret['comment'] += json.dumps(params, indent=4)

    ret['changes'] = {}
    # Return ``None`` when running with ``test=true``.
    ret['result'] = None

    return ret


def list_node(hostname, username, password, name):
    '''
    A function to connect to a bigip device and list a specific node.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    name
        The name of the node to list.
    '''

    ret = {'name': name, 'changes': {}, 'result': False, 'comment': ''}

    if __opts__['test']:
        return _test_output(ret, 'list', params={
            'hostname': hostname,
            'username': username,
            'password': password,
            'name': name
        }
        )

    response = __salt__['bigip.list_node'](hostname, username, password, name)
    return _load_result(response, ret)


def create_node(hostname, username, password, name, address):
    '''
    Create a new node if it does not already exist.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    name
        The name of the node to create
    address
        The address of the node
    '''

    ret = {'name': name, 'changes': {}, 'result': False, 'comment': ''}

    if __opts__['test']:
        return _test_output(ret, 'create', params={
            'hostname': hostname,
            'username': username,
            'password': password,
            'name': name,
            'address': address
        }
        )

    #is this node currently configured?
    existing = __salt__['bigip.list_node'](hostname, username, password, name)

    # if it exists
    if existing['code'] == 200:

        ret['result'] = True
        ret['comment'] = 'A node by this name currently exists.  No change made.'

    # if it doesn't exist
    elif existing['code'] == 404:
        response = __salt__['bigip.create_node'](hostname, username, password, name, address)

        ret['result'] = True
        ret['changes']['old'] = {}
        ret['changes']['new'] = response['content']
        ret['comment'] = 'Node was successfully created.'

    # else something else was returned
    else:
        ret = _load_result(existing, ret)

    return ret


def manage_node(hostname, username, password, name, address,
                connection_limit=None,
                description=None,
                dynamic_ratio=None,
                logging=None,
                monitor=None,
                rate_limit=None,
                ratio=None,
                session=None,
                node_state=None):
    '''
    Manages a node of a given bigip device.  If the node does not exist it will be created, otherwise,
    only the properties which are different than the existing will be updated.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    name
        The name of the node to manage.
    address
        The address of the node
    connection_limit
        [integer]
    description
        [string]
    dynam
        c_ratio:        [integer]
    logging
        [enabled | disabled]
    monitor
        [[name] | none | default]
    rate_limit
        [integer]
    ratio
        [integer]
    session
        [user-enabled | user-disabled]
    node_state (state)
        [user-down | user-up ]
    '''

    ret = {'name': name, 'changes': {}, 'result': False, 'comment': ''}

    if __opts__['test']:
        return _test_output(ret, 'manage', params={
            'hostname': hostname,
            'username': username,
            'password': password,
            'name': name,
            'address': address,
            'connection_limit': connection_limit,
            'description': description,
            'dynamic_ratio': dynamic_ratio,
            'logging': logging,
            'monitor': monitor,
            'rate_limit': rate_limit,
            'ratio': ratio,
            'session': session,
            'state:': node_state
        }
        )

    #is this node currently configured?
    existing = __salt__['bigip.list_node'](hostname, username, password, name)

    # if it exists by name
    if existing['code'] == 200:

        # ensure the address is the same, we don't want to modify a different node than what
        # we think we are managing
        if existing['content']['address'] != address:
            ret['result'] = False
            ret['comment'] = 'A node with this name exists but the address does not match.'

        modified = __salt__['bigip.modify_node'](hostname=hostname,
                                                 username=username,
                                                 password=password,
                                                 name=name,
                                                 connection_limit=connection_limit,
                                                 description=description,
                                                 dynamic_ratio=dynamic_ratio,
                                                 logging=logging,
                                                 monitor=monitor,
                                                 rate_limit=rate_limit,
                                                 ratio=ratio,
                                                 session=session,
                                                 state=node_state)

        #was the modification successful?
        if modified['code'] == 200:
            ret = _check_for_changes('Node', ret, existing, modified)
        else:
            ret = _load_result(modified, ret)

    # not found, attempt to create it
    elif existing['code'] == 404:

        new = __salt__['bigip.create_node'](hostname, username, password, name, address)

        # were we able to create it?
        if new['code'] == 200:
            # try modification

            modified = __salt__['bigip.modify_node'](hostname=hostname,
                                                     username=username,
                                                     password=password,
                                                     name=name,
                                                     connection_limit=connection_limit,
                                                     description=description,
                                                     dynamic_ratio=dynamic_ratio,
                                                     logging=logging,
                                                     monitor=monitor,
                                                     rate_limit=rate_limit,
                                                     ratio=ratio,
                                                     session=session,
                                                     state=node_state)
            #was the modification successful?
            if modified['code'] == 200:

                ret['result'] = True
                ret['comment'] = 'Node was created and enforced to the desired state.  Note: Only parameters specified ' \
                                 'were enforced.  See changes for details.'
                ret['changes']['old'] = {}
                ret['changes']['new'] = modified['content']

            # roll it back
            else:

                deleted = __salt__['bigip.delete_node'](hostname, username, password, name)
                # did we get rid of it?
                if deleted['code'] == 200:
                    ret['comment'] = 'Node was successfully created but an error occurred during modification. ' \
                                     'The creation of the node has been rolled back. Message is as follows:\n' \
                                     '{message}'.format(message=modified['content']['message'])
                # something bad happened
                else:
                    ret['comment'] = 'Node was successfully created but an error occurred during modification. ' \
                                     'The creation of the node was not able to be rolled back. Message is as follows:' \
                                     '\n {message}\n{message_two}'.format(message=modified['content']['message'],
                                                                          message_two=deleted['content']['message'])

        # unable to create it
        else:
            ret = _load_result(new, ret)
    # an error occurred
    else:
        ret = _load_result(existing, ret)

    return ret


def modify_node(hostname, username, password, name,
                connection_limit=None,
                description=None,
                dynamic_ratio=None,
                logging=None,
                monitor=None,
                rate_limit=None,
                ratio=None,
                session=None,
                node_state=None):
    '''
    Modify an existing node. Only a node which already exists will be modified and
    only the parameters specified will be enforced.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    name
        The name of the node to modify
    connection_limit
        [integer]
    description
        [string]
    dynamic_ratio
        [integer]
    logging
        [enabled | disabled]
    monitor
        [[name] | none | default]
    rate_limit
        [integer]
    ratio
        [integer]
    session
        [user-enabled | user-disabled]
    node_state (state)
        [user-down | user-up ]
    '''

    ret = {'name': name, 'changes': {}, 'result': False, 'comment': ''}

    if __opts__['test']:
        return _test_output(ret, 'modify', params={
            'hostname': hostname,
            'username': username,
            'password': password,
            'name': name,
            'connection_limit': connection_limit,
            'description': description,
            'dynamic_ratio': dynamic_ratio,
            'logging': logging,
            'monitor': monitor,
            'rate_limit': rate_limit,
            'ratio': ratio,
            'session': session,
            'state:': node_state
        }
        )

    #is this node currently configured?
    existing = __salt__['bigip.list_node'](hostname, username, password, name)

    # if it exists by name
    if existing['code'] == 200:

        modified = __salt__['bigip.modify_node'](hostname=hostname,
                                                 username=username,
                                                 password=password,
                                                 name=name,
                                                 connection_limit=connection_limit,
                                                 description=description,
                                                 dynamic_ratio=dynamic_ratio,
                                                 logging=logging,
                                                 monitor=monitor,
                                                 rate_limit=rate_limit,
                                                 ratio=ratio,
                                                 session=session,
                                                 state=node_state)

        #was the modification successful?
        if modified['code'] == 200:
            ret = _check_for_changes('Node', ret, existing, modified)
        else:
            ret = _load_result(modified, ret)

    # not found, attempt to create it
    elif existing['code'] == 404:
        ret['comment'] = 'A node with this name was not found.'
    # an error occurred
    else:
        ret = _load_result(existing, ret)

    return ret


def delete_node(hostname, username, password, name):
    '''
    Delete an existing node.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    name
        The name of the node which will be deleted.
    '''

    ret = {'name': name, 'changes': {}, 'result': False, 'comment': ''}

    if __opts__['test']:
        return _test_output(ret, 'delete', params={
            'hostname': hostname,
            'username': username,
            'password': password,
            'name': name,
        }
        )

    #is this node currently configured?
    existing = __salt__['bigip.list_node'](hostname, username, password, name)
    # if it exists by name
    if existing['code'] == 200:

        deleted = __salt__['bigip.delete_node'](hostname, username, password, name)
        # did we get rid of it?
        if deleted['code'] == 200:
            ret['result'] = True
            ret['comment'] = 'Node was successfully deleted.'
            ret['changes']['old'] = existing['content']
            ret['changes']['new'] = {}

        # something bad happened
        else:
            ret = _load_result(existing, ret)

    # not found
    elif existing['code'] == 404:
        ret['result'] = True
        ret['comment'] = 'This node already does not exist. No changes made.'
        ret['changes']['old'] = {}
        ret['changes']['new'] = {}

    else:
        ret = _load_result(existing, ret)

    return ret


def list_pool(hostname, username, password, name):
    '''
    A function to connect to a bigip device and list a specific pool.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    name
        The name of the pool to list.
    '''

    ret = {'name': name, 'changes': {}, 'result': False, 'comment': ''}

    if __opts__['test']:
        return _test_output(ret, 'list', params={
            'hostname': hostname,
            'username': username,
            'password': password,
            'name': name,
        }
        )

    response = __salt__['bigip.list_pool'](hostname, username, password, name)
    return _load_result(response, ret)


def create_pool(hostname, username, password, name, members=None,
                allow_nat=None,
                allow_snat=None,
                description=None,
                gateway_failsafe_device=None,
                ignore_persisted_weight=None,
                ip_tos_to_client=None,
                ip_tos_to_server=None,
                link_qos_to_client=None,
                link_qos_to_server=None,
                load_balancing_mode=None,
                min_active_members=None,
                min_up_members=None,
                min_up_members_action=None,
                min_up_members_checking=None,
                monitor=None,
                profiles=None,
                queue_depth_limit=None,
                queue_on_connection_limit=None,
                queue_time_limit=None,
                reselect_tries=None,
                service_down_action=None,
                slow_ramp_time=None):
    '''
    Create a new node if it does not already exist.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    name
        The name of the pool to create
    members
        List of members to be added to the pool
    allow_nat
        [yes | no]
    allow_snat
        [yes | no]
    description
        [string]
    gateway_failsafe_device
        [string]
    ignore_persisted_weight
        [enabled | disabled]
    ip_tos_to_client
        [pass-through | [integer]]
    ip_tos_to_server
        [pass-through | [integer]]
    link_qos_to_client
        [pass-through | [integer]]
    link_qos_to_server
        [pass-through | [integer]]
    load_balancing_mode
        [dynamic-ratio-member | dynamic-ratio-node |
        fastest-app-response | fastest-node |
        least-connections-members |
        least-connections-node |
        least-sessions |
        observed-member | observed-node |
        predictive-member | predictive-node |
        ratio-least-connections-member |
        ratio-least-connections-node |
        ratio-member | ratio-node | ratio-session |
        round-robin | weighted-least-connections-member |
        weighted-least-connections-node]
    min_active_members
        [integer]
    min_up_members
        [integer]
    min_up_members_action
        [failover | reboot | restart-all]
    min_up_members_checking
        [enabled | disabled]
    monitor
        [name]
    profiles
        [none | profile_name]
    queue_depth_limit
        [integer]
    queue_on_connection_limit
        [enabled | disabled]
    queue_time_limit
        [integer]
    reselect_tries
        [integer]
    service_down_action
        [drop | none | reselect | reset]
    slow_ramp_time
        [integer]
    '''

    ret = {'name': name, 'changes': {}, 'result': False, 'comment': ''}

    if __opts__['test']:
        return _test_output(ret, 'create', params={
            'hostname': hostname,
            'username': username,
            'password': password,
            'name': name,
            'members': members,
            'allow_nat': allow_nat,
            'allow_snat': allow_snat,
            'description': description,
            'gateway_failsafe_device': gateway_failsafe_device,
            'ignore_persisted_weight': ignore_persisted_weight,
            'ip_tos_client:': ip_tos_to_client,
            'ip_tos_server': ip_tos_to_server,
            'link_qos_to_client': link_qos_to_client,
            'link_qos_to_server': link_qos_to_server,
            'load_balancing_mode': load_balancing_mode,
            'min_active_members': min_active_members,
            'min_up_members': min_up_members,
            'min_up_members_checking': min_up_members_checking,
            'monitor': monitor,
            'profiles': profiles,
            'queue_depth_limit': queue_depth_limit,
            'queue_on_connection_limit': queue_on_connection_limit,
            'queue_time_limit': queue_time_limit,
            'reselect_tries': reselect_tries,
            'service_down_action': service_down_action,
            'slow_ramp_time': slow_ramp_time
            }
        )

    #is this pool currently configured?
    existing = __salt__['bigip.list_pool'](hostname, username, password, name)

    # if it exists
    if existing['code'] == 200:

        ret['result'] = True
        ret['comment'] = 'A pool by this name currently exists.  No change made.'

    # if it doesn't exist
    elif existing['code'] == 404:

        response = __salt__['bigip.create_pool'](hostname=hostname,
                                                 username=username,
                                                 password=password,
                                                 name=name,
                                                 members=members,
                                                 allow_nat=allow_nat,
                                                 allow_snat=allow_snat,
                                                 description=description,
                                                 gateway_failsafe_device=gateway_failsafe_device,
                                                 ignore_persisted_weight=ignore_persisted_weight,
                                                 ip_tos_to_client=ip_tos_to_client,
                                                 ip_tos_to_server=ip_tos_to_server,
                                                 link_qos_to_client=link_qos_to_client,
                                                 link_qos_to_server=link_qos_to_server,
                                                 load_balancing_mode=load_balancing_mode,
                                                 min_active_members=min_active_members,
                                                 min_up_members=min_up_members,
                                                 min_up_members_action=min_up_members_action,
                                                 min_up_members_checking=min_up_members_checking,
                                                 monitor=monitor,
                                                 profiles=profiles,
                                                 queue_depth_limit=queue_depth_limit,
                                                 queue_on_connection_limit=queue_on_connection_limit,
                                                 queue_time_limit=queue_time_limit,
                                                 reselect_tries=reselect_tries,
                                                 service_down_action=service_down_action,
                                                 slow_ramp_time=slow_ramp_time)
        if response['code'] == 200:
            ret['result'] = True
            ret['changes']['old'] = {}
            ret['changes']['new'] = response['content']
            ret['comment'] = 'Pool was successfully created.'
        else:
            ret = _load_result(existing, ret)

    # else something else was returned
    else:
        ret = _load_result(existing, ret)

    return ret


def manage_pool(hostname, username, password, name,
                allow_nat=None,
                allow_snat=None,
                description=None,
                gateway_failsafe_device=None,
                ignore_persisted_weight=None,
                ip_tos_to_client=None,
                ip_tos_to_server=None,
                link_qos_to_client=None,
                link_qos_to_server=None,
                load_balancing_mode=None,
                min_active_members=None,
                min_up_members=None,
                min_up_members_action=None,
                min_up_members_checking=None,
                monitor=None,
                profiles=None,
                queue_depth_limit=None,
                queue_on_connection_limit=None,
                queue_time_limit=None,
                reselect_tries=None,
                service_down_action=None,
                slow_ramp_time=None):
    '''
    Create a new pool if it does not already exist. Pool members are managed separately. Only the
    parameters specified are enforced.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    name
        The name of the pool to create
    allow_nat
        [yes | no]
    allow_snat
        [yes | no]
    description
        [string]
    gateway_failsafe_device
        [string]
    ignore_persisted_weight
        [enabled | disabled]
    ip_tos_to_client
        [pass-through | [integer]]
    ip_tos_to_server
        [pass-through | [integer]]
    link_qos_to_client
        [pass-through | [integer]]
    link_qos_to_server
        [pass-through | [integer]]
    load_balancing_mode
        [dynamic-ratio-member | dynamic-ratio-node |
        fastest-app-response | fastest-node |
        least-connections-members |
        least-connections-node |
        least-sessions |
        observed-member | observed-node |
        predictive-member | predictive-node |
        ratio-least-connections-member |
        ratio-least-connections-node |
        ratio-member | ratio-node | ratio-session |
        round-robin | weighted-least-connections-member |
        weighted-least-connections-node]
    min_active_members
        [integer]
    min_up_members
        [integer]
    min_up_members_action
        [failover | reboot | restart-all]
    min_up_members_checking
        [enabled | disabled]
    monitor
        [name]
    profiles
        [none | profile_name]
    queue_depth_limit
        [integer]
    queue_on_connection_limit
        [enabled | disabled]
    queue_time_limit
        [integer]
    reselect_tries
        [integer]
    service_down_action
        [drop | none | reselect | reset]
    slow_ramp_time
        [integer]

    '''

    ret = {'name': name, 'changes': {}, 'result': False, 'comment': ''}

    if __opts__['test']:
        return _test_output(ret, 'manage', params={
            'hostname': hostname,
            'username': username,
            'password': password,
            'name': name,
            'allow_nat': allow_nat,
            'allow_snat': allow_snat,
            'description': description,
            'gateway_failsafe_device': gateway_failsafe_device,
            'ignore_persisted_weight': ignore_persisted_weight,
            'ip_tos_client:': ip_tos_to_client,
            'ip_tos_server': ip_tos_to_server,
            'link_qos_to_client': link_qos_to_client,
            'link_qos_to_server': link_qos_to_server,
            'load_balancing_mode': load_balancing_mode,
            'min_active_members': min_active_members,
            'min_up_members': min_up_members,
            'min_up_members_checking': min_up_members_checking,
            'monitor': monitor,
            'profiles': profiles,
            'queue_depth_limit': queue_depth_limit,
            'queue_on_connection_limit': queue_on_connection_limit,
            'queue_time_limit': queue_time_limit,
            'reselect_tries': reselect_tries,
            'service_down_action': service_down_action,
            'slow_ramp_time': slow_ramp_time
        }
        )

    #is this pool currently configured?
    existing = __salt__['bigip.list_pool'](hostname, username, password, name)

    # if it exists
    if existing['code'] == 200:

        modified = __salt__['bigip.modify_pool'](hostname=hostname,
                                                 username=username,
                                                 password=password,
                                                 name=name,
                                                 allow_nat=allow_nat,
                                                 allow_snat=allow_snat,
                                                 description=description,
                                                 gateway_failsafe_device=gateway_failsafe_device,
                                                 ignore_persisted_weight=ignore_persisted_weight,
                                                 ip_tos_to_client=ip_tos_to_client,
                                                 ip_tos_to_server=ip_tos_to_server,
                                                 link_qos_to_client=link_qos_to_client,
                                                 link_qos_to_server=link_qos_to_server,
                                                 load_balancing_mode=load_balancing_mode,
                                                 min_active_members=min_active_members,
                                                 min_up_members=min_up_members,
                                                 min_up_members_action=min_up_members_action,
                                                 min_up_members_checking=min_up_members_checking,
                                                 monitor=monitor,
                                                 profiles=profiles,
                                                 queue_depth_limit=queue_depth_limit,
                                                 queue_on_connection_limit=queue_on_connection_limit,
                                                 queue_time_limit=queue_time_limit,
                                                 reselect_tries=reselect_tries,
                                                 service_down_action=service_down_action,
                                                 slow_ramp_time=slow_ramp_time)

        #was the modification successful?
        if modified['code'] == 200:

            #remove member listings and self-links
            del existing['content']['membersReference']
            del modified['content']['membersReference']
            del existing['content']['selfLink']
            del modified['content']['selfLink']

            ret = _check_for_changes('Pool', ret, existing, modified)
        else:
            ret = _load_result(modified, ret)

    # if it doesn't exist
    elif existing['code'] == 404:

        new = __salt__['bigip.create_pool'](hostname=hostname,
                                            username=username,
                                            password=password,
                                            name=name,
                                            allow_nat=allow_nat,
                                            allow_snat=allow_snat,
                                            description=description,
                                            gateway_failsafe_device=gateway_failsafe_device,
                                            ignore_persisted_weight=ignore_persisted_weight,
                                            ip_tos_to_client=ip_tos_to_client,
                                            ip_tos_to_server=ip_tos_to_server,
                                            link_qos_to_client=link_qos_to_client,
                                            link_qos_to_server=link_qos_to_server,
                                            load_balancing_mode=load_balancing_mode,
                                            min_active_members=min_active_members,
                                            min_up_members=min_up_members,
                                            min_up_members_action=min_up_members_action,
                                            min_up_members_checking=min_up_members_checking,
                                            monitor=monitor,
                                            profiles=profiles,
                                            queue_depth_limit=queue_depth_limit,
                                            queue_on_connection_limit=queue_on_connection_limit,
                                            queue_time_limit=queue_time_limit,
                                            reselect_tries=reselect_tries,
                                            service_down_action=service_down_action,
                                            slow_ramp_time=slow_ramp_time)

        # were we able to create it?
        if new['code'] == 200:
            ret['result'] = True
            ret['comment'] = 'Pool was created and enforced to the desired state.  Note: Only parameters specified ' \
                             'were enforced.  See changes for details.'
            ret['changes']['old'] = {}
            ret['changes']['new'] = new['content']

         # unable to create it
        else:
            ret = _load_result(new, ret)

    # else something else was returned
    else:
        ret = _load_result(existing, ret)

    return ret


def modify_pool(hostname, username, password, name,
                allow_nat=None,
                allow_snat=None,
                description=None,
                gateway_failsafe_device=None,
                ignore_persisted_weight=None,
                ip_tos_to_client=None,
                ip_tos_to_server=None,
                link_qos_to_client=None,
                link_qos_to_server=None,
                load_balancing_mode=None,
                min_active_members=None,
                min_up_members=None,
                min_up_members_action=None,
                min_up_members_checking=None,
                monitor=None,
                profiles=None,
                queue_depth_limit=None,
                queue_on_connection_limit=None,
                queue_time_limit=None,
                reselect_tries=None,
                service_down_action=None,
                slow_ramp_time=None):
    '''
    Modify an existing pool. Pool members are managed separately. Only the
    parameters specified are enforced.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    name
        The name of the pool to create
    allow_nat
        [yes | no]
    allow_snat
        [yes | no]
    description
        [string]
    gateway_failsafe_device
        [string]
    ignore_persisted_weight
        [enabled | disabled]
    ip_tos_to_client
        [pass-through | [integer]]
    ip_tos_to_server
        [pass-through | [integer]]
    link_qos_to_client
        [pass-through | [integer]]
    link_qos_to_server
        [pass-through | [integer]]
    load_balancing_mode
        [dynamic-ratio-member | dynamic-ratio-node |
        fastest-app-response | fastest-node |
        least-connections-members |
        least-connections-node |
        least-sessions |
        observed-member | observed-node |
        predictive-member | predictive-node |
        ratio-least-connections-member |
        ratio-least-connections-node |
        ratio-member | ratio-node | ratio-session |
        round-robin | weighted-least-connections-member |
        weighted-least-connections-node]
    min_active_members
        [integer]
    min_up_members
        [integer]
    min_up_members_action
        [failover | reboot | restart-all]
    min_up_members_checking
        [enabled | disabled]
    monitor
        [name]
    profiles
        [none | profile_name]
    queue_depth_limit
        [integer]
    queue_on_connection_limit
        [enabled | disabled]
    queue_time_limit
        [integer]
    reselect_tries
        [integer]
    service_down_action
        [drop | none | reselect | reset]
    slow_ramp_time
        [integer]

    '''

    ret = {'name': name, 'changes': {}, 'result': False, 'comment': ''}

    if __opts__['test']:
        return _test_output(ret, 'modify', params={
            'hostname': hostname,
            'username': username,
            'password': password,
            'name': name,
            'allow_nat': allow_nat,
            'allow_snat': allow_snat,
            'description': description,
            'gateway_failsafe_device': gateway_failsafe_device,
            'ignore_persisted_weight': ignore_persisted_weight,
            'ip_tos_client:': ip_tos_to_client,
            'ip_tos_server': ip_tos_to_server,
            'link_qos_to_client': link_qos_to_client,
            'link_qos_to_server': link_qos_to_server,
            'load_balancing_mode': load_balancing_mode,
            'min_active_members': min_active_members,
            'min_up_members': min_up_members,
            'min_up_members_checking': min_up_members_checking,
            'monitor': monitor,
            'profiles': profiles,
            'queue_depth_limit': queue_depth_limit,
            'queue_on_connection_limit': queue_on_connection_limit,
            'queue_time_limit': queue_time_limit,
            'reselect_tries': reselect_tries,
            'service_down_action': service_down_action,
            'slow_ramp_time': slow_ramp_time
        }
        )

    #is this pool currently configured?
    existing = __salt__['bigip.list_pool'](hostname, username, password, name)

    # if it exists
    if existing['code'] == 200:

        modified = __salt__['bigip.modify_pool'](hostname=hostname,
                                                 username=username,
                                                 password=password,
                                                 name=name,
                                                 allow_nat=allow_nat,
                                                 allow_snat=allow_snat,
                                                 description=description,
                                                 gateway_failsafe_device=gateway_failsafe_device,
                                                 ignore_persisted_weight=ignore_persisted_weight,
                                                 ip_tos_to_client=ip_tos_to_client,
                                                 ip_tos_to_server=ip_tos_to_server,
                                                 link_qos_to_client=link_qos_to_client,
                                                 link_qos_to_server=link_qos_to_server,
                                                 load_balancing_mode=load_balancing_mode,
                                                 min_active_members=min_active_members,
                                                 min_up_members=min_up_members,
                                                 min_up_members_action=min_up_members_action,
                                                 min_up_members_checking=min_up_members_checking,
                                                 monitor=monitor,
                                                 profiles=profiles,
                                                 queue_depth_limit=queue_depth_limit,
                                                 queue_on_connection_limit=queue_on_connection_limit,
                                                 queue_time_limit=queue_time_limit,
                                                 reselect_tries=reselect_tries,
                                                 service_down_action=service_down_action,
                                                 slow_ramp_time=slow_ramp_time)

        #was the modification successful?
        if modified['code'] == 200:

            #remove member listings and self-links
            del existing['content']['membersReference']
            del modified['content']['membersReference']
            del existing['content']['selfLink']
            del modified['content']['selfLink']

            ret = _check_for_changes('Pool', ret, existing, modified)
        else:
            ret = _load_result(modified, ret)

    # if it doesn't exist
    elif existing['code'] == 404:
        ret['comment'] = 'A pool with this name was not found.'
    # else something else was returned
    else:
        ret = _load_result(existing, ret)

    return ret


def delete_pool(hostname, username, password, name):
    '''
    Delete an existing pool.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    name
        The name of the pool which will be deleted
    '''

    ret = {'name': name, 'changes': {}, 'result': False, 'comment': ''}

    if __opts__['test']:
        return _test_output(ret, 'delete', params={
            'hostname': hostname,
            'username': username,
            'password': password,
            'name': name,
        }
        )

    #is this pool currently configured?
    existing = __salt__['bigip.list_pool'](hostname, username, password, name)
    # if it exists by name
    if existing['code'] == 200:

        deleted = __salt__['bigip.delete_pool'](hostname, username, password, name)
        # did we get rid of it?
        if deleted['code'] == 200:
            ret['result'] = True
            ret['comment'] = 'Pool was successfully deleted.'
            ret['changes']['old'] = existing['content']
            ret['changes']['new'] = {}

        # something bad happened
        else:
            ret = _load_result(deleted, ret)

    # not found
    elif existing['code'] == 404:
        ret['result'] = True
        ret['comment'] = 'This pool already does not exist. No changes made.'
        ret['changes']['old'] = {}
        ret['changes']['new'] = {}

    else:
        ret = _load_result(existing, ret)

    return ret


def manage_pool_members(hostname, username, password, name, members):
    '''
    Manage the members of an existing pool.  This function replaces all current pool members.
    Only the parameters specified are enforced.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    name
        The name of the pool to modify
    members
        list of pool members to manage.

    '''

    ret = {'name': name, 'changes': {}, 'result': False, 'comment': ''}

    if __opts__['test']:
        return _test_output(ret, 'manage', params={
            'hostname': hostname,
            'username': username,
            'password': password,
            'name': name,
            'members': members
        }
        )

    #is this pool currently configured?
    existing = __salt__['bigip.list_pool'](hostname, username, password, name)

    # if it exists
    if existing['code'] == 200:

        #what are the current members?
        current_members = existing['content']['membersReference']['items']

        modified = __salt__['bigip.replace_pool_members'](hostname, username, password, name, members)

        #was the modification successful?
        if modified['code'] == 200:

            #re-list the pool with new membership
            new_listing = __salt__['bigip.list_pool'](hostname, username, password, name)

            #just in case something happened...
            if new_listing['code'] != 200:
                ret = _load_result(new_listing, ret)
                ret['comment'] = 'modification of the pool was successful but an error occurred upon retrieving new' \
                                 ' listing.'
                return ret

            new_members = new_listing['content']['membersReference']['items']

            #remove generation keys and create new lists indexed by integers
            for current_member in current_members:
                del current_member['generation']

            for new_member in new_members:
                del new_member['generation']

            #anything changed?
            ret = _check_for_changes('Pool Membership', ret, current_members, new_members)

        else:
            ret = _load_result(modified, ret)

    #pool does not exists
    elif existing['code'] == 404:
        ret['comment'] = 'A pool with this name was not found.'

    else:
        ret = _load_result(existing, ret)

    return ret


def add_pool_member(hostname, username, password, name, member):
    '''
    A function to connect to a bigip device and add a new member to an existing pool.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    name
        The name of the pool to modify
    member
        The member to add to the pool
    '''

    ret = {'name': name, 'changes': {}, 'result': False, 'comment': ''}

    if __opts__['test']:
        return _test_output(ret, 'add', params={
            'hostname': hostname,
            'username': username,
            'password': password,
            'name': name,
            'members': member
        }
        )

    #is this pool member currently configured?
    existing_pool = __salt__['bigip.list_pool'](hostname, username, password, name)

    if existing_pool['code'] == 200:

        # for some reason iControl REST doesn't support listing a single pool member.
        # the response from GET for listing a member will return 200 even if it doesn't exists.
        # because of this we have to do some rather "unnecessary" searching within a pool.

        #what are the current members?
        current_members = existing_pool['content']['membersReference']['items']

        #loop through them
        exists = False
        for current_member in current_members:
            if current_member['name'] == member['name']:
                exists = True
                break

        if exists:
            ret['result'] = True
            ret['comment'] = 'Member: {name} already exists within this pool.  No changes made.'.format(name=member['name'])
            ret['changes']['old'] = {}
            ret['changes']['new'] = {}
        else:
            new_member = __salt__['bigip.add_pool_member'](hostname, username, password, name, member)

            if new_member['code'] == 200:
                ret['result'] = True
                ret['comment'] = 'Member: {name} has been successfully added to the pool.'.format(name=member['name'])
                ret['changes']['old'] = {}

                #look up the member again...
                pool_listing = __salt__['bigip.list_pool'](hostname, username, password, name)

                if pool_listing['code'] != 200:
                    ret = _load_result(new_member, ret)
                    return ret

                members = pool_listing['content']['membersReference']['items']
                #loop through them
                for current_member in members:
                    if current_member['name'] == member['name']:
                        added_member = current_member
                        break

                ret['changes']['new'] = added_member

            # member wasn't added
            else:
                ret = _load_result(new_member, ret)

    #pool does not exists
    elif existing_pool['code'] == 404:
        ret['comment'] = 'A pool with this name was not found.'
    else:
        ret = _load_result(existing_pool, ret)

    return ret


def modify_pool_member(hostname, username, password, name, member,
                       connection_limit=None,
                       description=None,
                       dynamic_ratio=None,
                       inherit_profile=None,
                       logging=None,
                       monitor=None,
                       priority_group=None,
                       profiles=None,
                       rate_limit=None,
                       ratio=None,
                       session=None,
                       member_state=None):
    '''
    A function to connect to a bigip device and modify a member of an existing pool.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    name
        The name of the pool to modify
    member
        The member modify
    connection_limit
        [integer]
    description
        [string]
    dynamic_ratio
        [integer]
    inherit_profile
        [enabled | disabled]
    logging
        [enabled | disabled]
    monitor
        [name]
    priority_group
        [integer]
    profiles
        [none | profile_name]
    rate_limit
        [integer]
    ratio
        [integer]
    session
        [user-enabled | user-disabled]
    member_state (state)
        [ user-up | user-down ]

    '''

    ret = {'name': name, 'changes': {}, 'result': False, 'comment': ''}

    if __opts__['test']:
        return _test_output(ret, 'modify', params={
            'hostname': hostname,
            'username': username,
            'password': password,
            'name': name,
            'members': member
        }
        )

    #is this pool member currently configured?
    existing_pool = __salt__['bigip.list_pool'](hostname, username, password, name)

    if existing_pool['code'] == 200:

        # for some reason iControl REST doesn't support listing a single pool member.
        # the response from GET for listing a member will return 200 even if it doesn't exists.
        # because of this we have to do some rather "unnecessary" searching within a pool.

        #what are the current members?
        current_members = existing_pool['content']['membersReference']['items']

        #loop through them
        exists = False
        for current_member in current_members:
            if current_member['name'] == member:
                exists = True
                existing_member = current_member
                break

        if exists:

            #modify the pool member
            modified = __salt__['bigip.modify_pool_member'](hostname=hostname,
                                                            username=username,
                                                            password=password,
                                                            name=name,
                                                            member=member,
                                                            connection_limit=connection_limit,
                                                            description=description,
                                                            dynamic_ratio=dynamic_ratio,
                                                            inherit_profile=inherit_profile,
                                                            logging=logging,
                                                            monitor=monitor,
                                                            priority_group=priority_group,
                                                            profiles=profiles,
                                                            rate_limit=rate_limit,
                                                            ratio=ratio,
                                                            session=session,
                                                            state=member_state)

            #re-list the pool
            new_pool = __salt__['bigip.list_pool'](hostname, username, password, name)

            if modified['code'] == 200 and modified['code'] == 200:

                #what are the new members?
                new_members = new_pool['content']['membersReference']['items']

                #loop through them
                for new_member in new_members:
                    if new_member['name'] == member:
                        modified_member = new_member
                        break

                #check for changes
                old = {'content': existing_member}
                new = {'content': modified_member}
                ret = _check_for_changes('Pool Member: {member}'.format(member=member), ret, old, new)

            else:
                ret = _load_result(modified, ret)
        else:
            ret['comment'] = 'Member: {name} does not exists within this pool.  No changes made.'.format(name=member['name'])

    #pool does not exists
    elif existing_pool['code'] == 404:
        ret['comment'] = 'A pool with this name was not found.'
    else:
        ret = _load_result(existing_pool, ret)

    return ret


def delete_pool_member(hostname, username, password, name, member):
    '''
    Delete an existing pool member.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    name
        The name of the pool to be modified
    member
        The name of the member to delete from the pool
    '''

    ret = {'name': name, 'changes': {}, 'result': False, 'comment': ''}

    if __opts__['test']:
        return _test_output(ret, 'delete', params={
            'hostname': hostname,
            'username': username,
            'password': password,
            'name': name,
            'members': member
        }
        )

    #is this pool currently configured?
    existing = __salt__['bigip.list_pool'](hostname, username, password, name)

    # if it exists by name
    if existing['code'] == 200:

       #what are the current members?
        current_members = existing['content']['membersReference']['items']

        #loop through them
        exists = False
        for current_member in current_members:
            if current_member['name'] == member:
                exists = True
                existing_member = current_member
                break

        if exists:
            deleted = __salt__['bigip.delete_pool_member'](hostname, username, password, name, member)
            # did we get rid of it?
            if deleted['code'] == 200:
                ret['result'] = True
                ret['comment'] = 'Pool Member: {member} was successfully deleted.'.format(member=member)
                ret['changes']['old'] = existing_member
                ret['changes']['new'] = {}

        # something bad happened
        else:
            ret['result'] = True
            ret['comment'] = 'This pool member already does not exist. No changes made.'
            ret['changes']['old'] = {}
            ret['changes']['new'] = {}

    else:
        ret = _load_result(existing, ret)

    return ret


def list_virtual(hostname, username, password, name):
    '''
    A function to list a specific virtual.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    name
        The name of the virtual to list
    '''

    ret = {'name': name, 'changes': {}, 'result': False, 'comment': ''}

    if __opts__['test']:
        return _test_output(ret, 'list', params={
            'hostname': hostname,
            'username': username,
            'password': password,
            'name': name
        }
        )

    response = __salt__['bigip.list_virtual'](hostname, username, password, name)
    return _load_result(response, ret)


def create_virtual(hostname, username, password, name, destination,
                   pool=None,
                   address_status=None,
                   auto_lasthop=None,
                   bwc_policy=None,
                   cmp_enabled=None,
                   connection_limit=None,
                   dhcp_relay=None,
                   description=None,
                   fallback_persistence=None,
                   flow_eviction_policy=None,
                   gtm_score=None,
                   ip_forward=None,
                   ip_protocol=None,
                   internal=None,
                   twelve_forward=None,
                   last_hop_pool=None,
                   mask=None,
                   mirror=None,
                   nat64=None,
                   persist=None,
                   profiles=None,
                   policies=None,
                   rate_class=None,
                   rate_limit=None,
                   rate_limit_mode=None,
                   rate_limit_dst=None,
                   rate_limit_src=None,
                   rules=None,
                   related_rules=None,
                   reject=None,
                   source=None,
                   source_address_translation=None,
                   source_port=None,
                   virtual_state=None,
                   traffic_classes=None,
                   translate_address=None,
                   translate_port=None,
                   vlans=None):
    '''
    A function to connect to a bigip device and create a virtual server if it does not already exists.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    name
        The name of the virtual to create
    destination
        [ [virtual_address_name:port] | [ipv4:port] | [ipv6.port] ]
    pool
        [ [pool_name] | none]
    address_status
        [yes | no]
    auto_lasthop
        [default | enabled | disabled ]
    bwc_policy
        [none] | string]
    cmp_enabled
        [yes | no]
    dhcp_relay
        [yes | no}
    connection_limit
        [integer]
    description
        [string]
    state
        [disabled | enabled]
    fallback_persistence
        [none | [profile name] ]
    flow_eviction_policy
        [none | [eviction policy name] ]
    gtm_score
        [integer]
    ip_forward
        [yes | no]
    ip_protocol
        [any | protocol]
    internal
        [yes | no]
    twelve_forward(12-forward)
        [yes | no]
    last_hop-pool
        [ [pool_name] | none]
    mask
        { [ipv4] | [ipv6] }
    mirror
        { [disabled | enabled | none] }
    nat64
        [enabled | disabled]
    persist
        [list]
    profiles
        [none | default | list ]
    policies
        [none | default | list ]
    rate_class
        [name]
    rate_limit
        [integer]
    rate_limit-mode
        [destination | object | object-destination |
        object-source | object-source-destination |
        source | source-destination]
    rate_limit-dst
        [integer]
    rate_limit-src
        [integer]
    rules
        [none | list ]
    related_rules
        [none | list ]
    reject
        [yes | no]
    source
        { [ipv4[/prefixlen]] | [ipv6[/prefixlen]] }
    source_address_translation
        [none | snat:pool_name | lsn | automap | dictionary ]
    source_port
        [change | preserve | preserve-strict]
    state
        [enabled | disabled]
    traffic_classes
        [none | default | list ]
    translate_address
        [enabled | disabled]
    translate_port
        [enabled | disabled]
    vlans
        [none | default | dictionary]

        vlan_ids
            [ list]
        enabled
            [ true | false ]
    '''

    ret = {'name': name, 'changes': {}, 'result': False, 'comment': ''}

    if __opts__['test']:
        return _test_output(ret, 'create', params={
            'hostname': hostname,
            'username': username,
            'password': password,
            'name': name,
            'destination': destination,
            'pool': pool,
            'address_status': address_status,
            'auto_lasthop': auto_lasthop,
            'bwc_policy': bwc_policy,
            'cmp_enabled': cmp_enabled,
            'connection_limit': connection_limit,
            'dhcp_relay': dhcp_relay,
            'description': description,
            'fallback_persistence': fallback_persistence,
            'flow_eviction_policy': flow_eviction_policy,
            'gtm_score': gtm_score,
            'ip_forward': ip_forward,
            'ip_protocol': ip_protocol,
            'internal': internal,
            'twelve_forward': twelve_forward,
            'last_hop_pool': last_hop_pool,
            'mask': mask,
            'mirror': mirror,
            'nat64': nat64,
            'persist': persist,
            'profiles': profiles,
            'policies': policies,
            'rate_class': rate_class,
            'rate_limit': rate_limit,
            'rate_limit_mode': rate_limit_mode,
            'rate_limit_dst': rate_limit_dst,
            'rate_limit_src': rate_limit_src,
            'rules': rules,
            'related_rules': related_rules,
            'reject': reject,
            'source': source,
            'source_address_translation': source_address_translation,
            'source_port': source_port,
            'virtual_state': virtual_state,
            'traffic_classes': traffic_classes,
            'translate_address': translate_address,
            'translate_port': translate_port,
            'vlans': vlans
        }
        )

    existing = __salt__['bigip.list_virtual'](hostname, username, password, name)

    # does this virtual exist?
    if existing['code'] == 200:

        ret['result'] = True
        ret['comment'] = 'A virtual by this name currently exists.  No change made.'

    elif existing['code'] == 404:

        #create it
        virtual = __salt__['bigip.create_virtual'](hostname=hostname,
                                                   username=username,
                                                   password=password,
                                                   name=name,
                                                   destination=destination,
                                                   description=description,
                                                   pool=pool,
                                                   address_status=address_status,
                                                   auto_lasthop=auto_lasthop,
                                                   bwc_policy=bwc_policy,
                                                   cmp_enabled=cmp_enabled,
                                                   connection_limit=connection_limit,
                                                   dhcp_relay=dhcp_relay,
                                                   fallback_persistence=fallback_persistence,
                                                   flow_eviction_policy=flow_eviction_policy,
                                                   gtm_score=gtm_score,
                                                   ip_forward=ip_forward,
                                                   ip_protocol=ip_protocol,
                                                   internal=internal,
                                                   twelve_forward=twelve_forward,
                                                   last_hop_pool=last_hop_pool,
                                                   mask=mask,
                                                   mirror=mirror,
                                                   nat64=nat64,
                                                   persist=persist,
                                                   profiles=profiles,
                                                   policies=policies,
                                                   rate_class=rate_class,
                                                   rate_limit=rate_limit,
                                                   rate_limit_mode=rate_limit_mode,
                                                   rate_limit_dst=rate_limit_dst,
                                                   rate_limit_src=rate_limit_src,
                                                   rules=rules,
                                                   related_rules=related_rules,
                                                   reject=reject,
                                                   source=source,
                                                   source_address_translation=source_address_translation,
                                                   source_port=source_port,
                                                   state=virtual_state,
                                                   traffic_classes=traffic_classes,
                                                   translate_address=translate_address,
                                                   translate_port=translate_port,
                                                   vlans=vlans)
        if virtual['code'] == 200:
            ret['result'] = True
            ret['changes']['old'] = {}
            ret['changes']['new'] = virtual['content']
            ret['comment'] = 'Virtual was successfully created.'
        else:
            ret = _load_result(existing, ret)

    # else something else was returned
    else:
        ret = _load_result(existing, ret)

    return ret


def manage_virtual(hostname, username, password, name, destination,
                   pool=None,
                   address_status=None,
                   auto_lasthop=None,
                   bwc_policy=None,
                   cmp_enabled=None,
                   connection_limit=None,
                   dhcp_relay=None,
                   description=None,
                   fallback_persistence=None,
                   flow_eviction_policy=None,
                   gtm_score=None,
                   ip_forward=None,
                   ip_protocol=None,
                   internal=None,
                   twelve_forward=None,
                   last_hop_pool=None,
                   mask=None,
                   mirror=None,
                   nat64=None,
                   persist=None,
                   profiles=None,
                   policies=None,
                   rate_class=None,
                   rate_limit=None,
                   rate_limit_mode=None,
                   rate_limit_dst=None,
                   rate_limit_src=None,
                   rules=None,
                   related_rules=None,
                   reject=None,
                   source=None,
                   source_address_translation=None,
                   source_port=None,
                   virtual_state=None,
                   traffic_classes=None,
                   translate_address=None,
                   translate_port=None,
                   vlans=None):
    '''
    Manage a virtual server.  If a virtual does not exists it will be created, otherwise only the
    parameters specified will be enforced.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    name
        The name of the virtual to create
    destination
        [ [virtual_address_name:port] | [ipv4:port] | [ipv6.port] ]
    pool
        [ [pool_name] | none]
    address_status
        [yes | no]
    auto_lasthop
        [default | enabled | disabled ]
    bwc_policy
        [none] | string]
    cmp_enabled
        [yes | no]
    dhcp_relay
        [yes | no}
    connection_limit
        [integer]
    description
        [string]
    state
        [disabled | enabled]
    fallback_persistence
        [none | [profile name] ]
    flow_eviction_policy
        [none | [eviction policy name] ]
    gtm_score
        [integer]
    ip_forward
        [yes | no]
    ip_protocol
        [any | protocol]
    internal
        [yes | no]
    twelve_forward(12-forward)
        [yes | no]
    last_hop-pool
        [ [pool_name] | none]
    mask
        { [ipv4] | [ipv6] }
    mirror
        { [disabled | enabled | none] }
    nat64
        [enabled | disabled]
    persist
        [list]
    profiles
        [none | default | list ]
    policies
        [none | default | list ]
    rate_class
        [name]
    rate_limit
        [integer]
    rate_limit-mode
        [destination | object | object-destination |
        object-source | object-source-destination |
        source | source-destination]
    rate_limit-dst
        [integer]
    rate_limit-src
        [integer]
    rules
        [none | list ]
    related_rules
        [none | list ]
    reject
        [yes | no]
    source
        { [ipv4[/prefixlen]] | [ipv6[/prefixlen]] }
    source_address_translation
        [none | snat:pool_name | lsn | automap | dictionary ]
    source_port
        [change | preserve | preserve-strict]
    state
        [enabled | disabled]
    traffic_classes
        [none | default | list ]
    translate_address
        [enabled | disabled]
    translate_port
        [enabled | disabled]
    vlans
        [none | default | dictionary]

        vlan_ids
            [ list]
        enabled
            [ true | false ]
    '''

    ret = {'name': name, 'changes': {}, 'result': False, 'comment': ''}

    if __opts__['test']:
        return _test_output(ret, 'manage', params={
            'hostname': hostname,
            'username': username,
            'password': password,
            'name': name,
            'destination': destination,
            'pool': pool,
            'address_status': address_status,
            'auto_lasthop': auto_lasthop,
            'bwc_policy': bwc_policy,
            'cmp_enabled': cmp_enabled,
            'connection_limit': connection_limit,
            'dhcp_relay': dhcp_relay,
            'description': description,
            'fallback_persistence': fallback_persistence,
            'flow_eviction_policy': flow_eviction_policy,
            'gtm_score': gtm_score,
            'ip_forward': ip_forward,
            'ip_protocol': ip_protocol,
            'internal': internal,
            'twelve_forward': twelve_forward,
            'last_hop_pool': last_hop_pool,
            'mask': mask,
            'mirror': mirror,
            'nat64': nat64,
            'persist': persist,
            'profiles': profiles,
            'policies': policies,
            'rate_class': rate_class,
            'rate_limit': rate_limit,
            'rate_limit_mode': rate_limit_mode,
            'rate_limit_dst': rate_limit_dst,
            'rate_limit_src': rate_limit_src,
            'rules': rules,
            'related_rules': related_rules,
            'reject': reject,
            'source': source,
            'source_address_translation': source_address_translation,
            'source_port': source_port,
            'virtual_state': virtual_state,
            'traffic_classes': traffic_classes,
            'translate_address': translate_address,
            'translate_port': translate_port,
            'vlans': vlans
        }
        )

    existing = __salt__['bigip.list_virtual'](hostname, username, password, name)

    # does this virtual exist?
    if existing['code'] == 200:

        # modify
        modified = __salt__['bigip.modify_virtual'](hostname=hostname,
                                                    username=username,
                                                    password=password,
                                                    name=name,
                                                    destination=destination,
                                                    description=description,
                                                    pool=pool,
                                                    address_status=address_status,
                                                    auto_lasthop=auto_lasthop,
                                                    bwc_policy=bwc_policy,
                                                    cmp_enabled=cmp_enabled,
                                                    connection_limit=connection_limit,
                                                    dhcp_relay=dhcp_relay,
                                                    fallback_persistence=fallback_persistence,
                                                    flow_eviction_policy=flow_eviction_policy,
                                                    gtm_score=gtm_score,
                                                    ip_forward=ip_forward,
                                                    ip_protocol=ip_protocol,
                                                    internal=internal,
                                                    twelve_forward=twelve_forward,
                                                    last_hop_pool=last_hop_pool,
                                                    mask=mask,
                                                    mirror=mirror,
                                                    nat64=nat64,
                                                    persist=persist,
                                                    profiles=profiles,
                                                    policies=policies,
                                                    rate_class=rate_class,
                                                    rate_limit=rate_limit,
                                                    rate_limit_mode=rate_limit_mode,
                                                    rate_limit_dst=rate_limit_dst,
                                                    rate_limit_src=rate_limit_src,
                                                    rules=rules,
                                                    related_rules=related_rules,
                                                    reject=reject,
                                                    source=source,
                                                    source_address_translation=source_address_translation,
                                                    source_port=source_port,
                                                    state=virtual_state,
                                                    traffic_classes=traffic_classes,
                                                    translate_address=translate_address,
                                                    translate_port=translate_port,
                                                    vlans=vlans)

        #was the modification successful?
        if modified['code'] == 200:

            #relist it to compare
            relisting = __salt__['bigip.list_virtual'](hostname, username, password, name)

            if relisting['code'] == 200:

                relisting = _strip_key(relisting, 'generation')
                existing = _strip_key(existing, 'generation')

                ret = _check_for_changes('Virtual', ret, existing, relisting)
            else:
                ret = _load_result(relisting, ret)

        else:
            ret = _load_result(modified, ret)

    elif existing['code'] == 404:

        #create it
        virtual = __salt__['bigip.create_virtual'](hostname=hostname,
                                                   username=username,
                                                   password=password,
                                                   name=name,
                                                   destination=destination,
                                                   description=description,
                                                   pool=pool,
                                                   address_status=address_status,
                                                   auto_lasthop=auto_lasthop,
                                                   bwc_policy=bwc_policy,
                                                   cmp_enabled=cmp_enabled,
                                                   connection_limit=connection_limit,
                                                   dhcp_relay=dhcp_relay,
                                                   fallback_persistence=fallback_persistence,
                                                   flow_eviction_policy=flow_eviction_policy,
                                                   gtm_score=gtm_score,
                                                   ip_forward=ip_forward,
                                                   ip_protocol=ip_protocol,
                                                   internal=internal,
                                                   twelve_forward=twelve_forward,
                                                   last_hop_pool=last_hop_pool,
                                                   mask=mask,
                                                   mirror=mirror,
                                                   nat64=nat64,
                                                   persist=persist,
                                                   profiles=profiles,
                                                   policies=policies,
                                                   rate_class=rate_class,
                                                   rate_limit=rate_limit,
                                                   rate_limit_mode=rate_limit_mode,
                                                   rate_limit_dst=rate_limit_dst,
                                                   rate_limit_src=rate_limit_src,
                                                   rules=rules,
                                                   related_rules=related_rules,
                                                   reject=reject,
                                                   source=source,
                                                   source_address_translation=source_address_translation,
                                                   source_port=source_port,
                                                   state=virtual_state,
                                                   traffic_classes=traffic_classes,
                                                   translate_address=translate_address,
                                                   translate_port=translate_port,
                                                   vlans=vlans)

        #were we able to create it?
        if virtual['code'] == 200:
            ret['result'] = True
            ret['changes']['old'] = {}
            ret['changes']['new'] = virtual['content']
            ret['comment'] = 'Virtual was successfully created and enforced to the desired state.'

        else:
            ret = _load_result(virtual, ret)

    else:
        ret = _load_result(existing, ret)

    return ret


def modify_virtual(hostname, username, password, name, destination,
                   pool=None,
                   address_status=None,
                   auto_lasthop=None,
                   bwc_policy=None,
                   cmp_enabled=None,
                   connection_limit=None,
                   dhcp_relay=None,
                   description=None,
                   fallback_persistence=None,
                   flow_eviction_policy=None,
                   gtm_score=None,
                   ip_forward=None,
                   ip_protocol=None,
                   internal=None,
                   twelve_forward=None,
                   last_hop_pool=None,
                   mask=None,
                   mirror=None,
                   nat64=None,
                   persist=None,
                   profiles=None,
                   policies=None,
                   rate_class=None,
                   rate_limit=None,
                   rate_limit_mode=None,
                   rate_limit_dst=None,
                   rate_limit_src=None,
                   rules=None,
                   related_rules=None,
                   reject=None,
                   source=None,
                   source_address_translation=None,
                   source_port=None,
                   virtual_state=None,
                   traffic_classes=None,
                   translate_address=None,
                   translate_port=None,
                   vlans=None):
    '''
    Modify an virtual server.  modify an existing virtual.  Only parameters specified will be enforced.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    name
        The name of the virtual to create
    destination
        [ [virtual_address_name:port] | [ipv4:port] | [ipv6.port] ]
    pool
        [ [pool_name] | none]
    address_status
        [yes | no]
    auto_lasthop
        [default | enabled | disabled ]
    bwc_policy
        [none] | string]
    cmp_enabled
        [yes | no]
    dhcp_relay
        [yes | no}
    connection_limit
        [integer]
    description
        [string]
    state
        [disabled | enabled]
    fallback_persistence
        [none | [profile name] ]
    flow_eviction_policy
        [none | [eviction policy name] ]
    gtm_score
        [integer]
    ip_forward
        [yes | no]
    ip_protocol
        [any | protocol]
    internal
        [yes | no]
    twelve_forward(12-forward)
        [yes | no]
    last_hop-pool
        [ [pool_name] | none]
    mask
        { [ipv4] | [ipv6] }
    mirror
        { [disabled | enabled | none] }
    nat64
        [enabled | disabled]
    persist
        [list]
    profiles
        [none | default | list ]
    policies
        [none | default | list ]
    rate_class
        [name]
    rate_limit
        [integer]
    rate_limit-mode
        [destination | object | object-destination |
        object-source | object-source-destination |
        source | source-destination]
    rate_limit_dst
        [integer]
    rate_limit_src
        [integer]
    rules
        [none | list ]
    related_rules
        [none | list ]
    reject
        [yes | no]
    source
        { [ipv4[/prefixlen]] | [ipv6[/prefixlen]] }
    source_address_translation
        [none | snat:pool_name | lsn | automap | dictionary ]
    source_port
        [change | preserve | preserve-strict]
    state
        [enabled | disabled]
    traffic_classes
        [none | default | list ]
    translate_address
        [enabled | disabled]
    translate_port
        [enabled | disabled]
    vlans
        [none | default | dictionary ]

        vlan_ids
            [ list]
        enabled
            [ true | false ]

    '''

    ret = {'name': name, 'changes': {}, 'result': False, 'comment': ''}

    if __opts__['test']:
        return _test_output(ret, 'modify', params={
            'hostname': hostname,
            'username': username,
            'password': password,
            'name': name,
            'destination': destination,
            'pool': pool,
            'address_status': address_status,
            'auto_lasthop': auto_lasthop,
            'bwc_policy': bwc_policy,
            'cmp_enabled': cmp_enabled,
            'connection_limit': connection_limit,
            'dhcp_relay': dhcp_relay,
            'description': description,
            'fallback_persistence': fallback_persistence,
            'flow_eviction_policy': flow_eviction_policy,
            'gtm_score': gtm_score,
            'ip_forward': ip_forward,
            'ip_protocol': ip_protocol,
            'internal': internal,
            'twelve_forward': twelve_forward,
            'last_hop_pool': last_hop_pool,
            'mask': mask,
            'mirror': mirror,
            'nat64': nat64,
            'persist': persist,
            'profiles': profiles,
            'policies': policies,
            'rate_class': rate_class,
            'rate_limit': rate_limit,
            'rate_limit_mode': rate_limit_mode,
            'rate_limit_dst': rate_limit_dst,
            'rate_limit_src': rate_limit_src,
            'rules': rules,
            'related_rules': related_rules,
            'reject': reject,
            'source': source,
            'source_address_translation': source_address_translation,
            'source_port': source_port,
            'virtual_state': virtual_state,
            'traffic_classes': traffic_classes,
            'translate_address': translate_address,
            'translate_port': translate_port,
            'vlans': vlans
        }
        )

    existing = __salt__['bigip.list_virtual'](hostname, username, password, name)

    # does this virtual exist?
    if existing['code'] == 200:

        # modify
        modified = __salt__['bigip.modify_virtual'](hostname=hostname,
                                                    username=username,
                                                    password=password,
                                                    name=name,
                                                    destination=destination,
                                                    description=description,
                                                    pool=pool,
                                                    address_status=address_status,
                                                    auto_lasthop=auto_lasthop,
                                                    bwc_policy=bwc_policy,
                                                    cmp_enabled=cmp_enabled,
                                                    connection_limit=connection_limit,
                                                    dhcp_relay=dhcp_relay,
                                                    fallback_persistence=fallback_persistence,
                                                    flow_eviction_policy=flow_eviction_policy,
                                                    gtm_score=gtm_score,
                                                    ip_forward=ip_forward,
                                                    ip_protocol=ip_protocol,
                                                    internal=internal,
                                                    twelve_forward=twelve_forward,
                                                    last_hop_pool=last_hop_pool,
                                                    mask=mask,
                                                    mirror=mirror,
                                                    nat64=nat64,
                                                    persist=persist,
                                                    profiles=profiles,
                                                    policies=policies,
                                                    rate_class=rate_class,
                                                    rate_limit=rate_limit,
                                                    rate_limit_mode=rate_limit_mode,
                                                    rate_limit_dst=rate_limit_dst,
                                                    rate_limit_src=rate_limit_src,
                                                    rules=rules,
                                                    related_rules=related_rules,
                                                    reject=reject,
                                                    source=source,
                                                    source_address_translation=source_address_translation,
                                                    source_port=source_port,
                                                    state=virtual_state,
                                                    traffic_classes=traffic_classes,
                                                    translate_address=translate_address,
                                                    translate_port=translate_port,
                                                    vlans=vlans)

        #was the modification successful?
        if modified['code'] == 200:

            #relist it to compare
            relisting = __salt__['bigip.list_virtual'](hostname, username, password, name)

            if relisting['code'] == 200:

                relisting = _strip_key(relisting, 'generation')
                existing = _strip_key(existing, 'generation')

                ret = _check_for_changes('Virtual', ret, existing, relisting)
            else:
                ret = _load_result(relisting, ret)

        else:
            ret = _load_result(modified, ret)

    elif existing['code'] == 404:
        ret['comment'] = 'A Virtual with this name was not found.'
        # else something else was returned
    else:
        ret = _load_result(existing, ret)

    return ret


def delete_virtual(hostname, username, password, name):
    '''
    Delete an existing virtual.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    name
        The name of the virtual which will be deleted
    '''

    ret = {'name': name, 'changes': {}, 'result': False, 'comment': ''}

    if __opts__['test']:
        return _test_output(ret, 'delete', params={
            'hostname': hostname,
            'username': username,
            'password': password,
            'name': name
        }
        )

    #is this virtual currently configured?
    existing = __salt__['bigip.list_virtual'](hostname, username, password, name)
    # if it exists by name
    if existing['code'] == 200:

        deleted = __salt__['bigip.delete_virtual'](hostname, username, password, name)
        # did we get rid of it?
        if deleted['code'] == 200:
            ret['result'] = True
            ret['comment'] = 'Virtual was successfully deleted.'
            ret['changes']['old'] = existing['content']
            ret['changes']['new'] = {}
        # something bad happened
        else:
            ret = _load_result(deleted, ret)

    # not found
    elif existing['code'] == 404:
        ret['result'] = True
        ret['comment'] = 'This virtual already does not exist. No changes made.'
        ret['changes']['old'] = {}
        ret['changes']['new'] = {}
    else:
        ret = _load_result(existing, ret)

    return ret


def list_monitor(hostname, username, password, monitor_type, name):
    '''
    A function to list an exsiting monitor.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    monitor_type
        The type of monitor to list
    name
        The name of the monitor to list
    '''

    ret = {'name': name, 'changes': {}, 'result': False, 'comment': ''}

    if __opts__['test']:
        return _test_output(ret, 'list', params={
            'hostname': hostname,
            'username': username,
            'password': password,
            'monitor_type': monitor_type,
            'name': name
        }
        )

    response = __salt__['bigip.list_monitor'](hostname, username, password, monitor_type, name)
    return _load_result(response, ret)


def create_monitor(hostname, username, password, monitor_type, name, **kwargs):
    '''
    A function to connect to a bigip device and create a monitor.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    monitor_type
        The type of monitor to create
    name
        The name of the monitor to create
    kwargs
        [ arg=val ] ...

        Consult F5 BIGIP user guide for specific options for each monitor type.
        Typically, tmsh arg names are used.
    '''

    ret = {'name': name, 'changes': {}, 'result': False, 'comment': ''}

    if __opts__['test']:

        params = {
            'hostname': hostname,
            'username': username,
            'password': password,
            'monitor_type': monitor_type,
            'name': name
        }

        for key, value in six.iteritems(kwargs):
            params[key] = value

        return _test_output(ret, 'create', params)

    #is this monitor currently configured?
    existing = __salt__['bigip.list_monitor'](hostname, username, password, monitor_type, name)

    # if it exists
    if existing['code'] == 200:

        ret['result'] = True
        ret['comment'] = 'A monitor by this name currently exists.  No change made.'

    # if it doesn't exist
    elif existing['code'] == 404:

        response = __salt__['bigip.create_monitor'](hostname, username, password, monitor_type, name, **kwargs)
        if response['code'] == 200:
            ret['result'] = True
            ret['changes']['old'] = {}
            ret['changes']['new'] = response['content']
            ret['comment'] = 'Monitor was successfully created.'
        else:
            ret = _load_result(response, ret)

    # else something else was returned
    else:
        ret = _load_result(existing, ret)

    return ret


def manage_monitor(hostname, username, password, monitor_type, name, **kwargs):
    '''
    Create a new monitor if a monitor of this type and name does not already exists.  If it does exists, only
    the parameters specified will be enforced.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    monitor_type
        The type of monitor to create
    name
        The name of the monitor to create
    kwargs
        [ arg=val ] ...

        Consult F5 BIGIP user guide for specific options for each monitor type.
        Typically, tmsh arg names are used.
    '''

    ret = {'name': name, 'changes': {}, 'result': False, 'comment': ''}

    if __opts__['test']:

        params = {
            'hostname': hostname,
            'username': username,
            'password': password,
            'monitor_type': monitor_type,
            'name': name
        }

        for key, value in six.iteritems(kwargs):
            params[key] = value

        return _test_output(ret, 'manage', params)

    #is this monitor currently configured?
    existing = __salt__['bigip.list_monitor'](hostname, username, password, monitor_type, name)

    # if it exists
    if existing['code'] == 200:

        #modify the monitor
        modified = __salt__['bigip.modify_monitor'](hostname, username, password, monitor_type, name, **kwargs)

        #was the modification successful?
        if modified['code'] == 200:
            del existing['content']['selfLink']
            del modified['content']['selfLink']
            ret = _check_for_changes('Monitor', ret, existing, modified)
        else:
            ret = _load_result(modified, ret)

    # if it doesn't exist
    elif existing['code'] == 404:

        response = __salt__['bigip.create_monitor'](hostname, username, password, monitor_type, name, **kwargs)
        if response['code'] == 200:
            ret['result'] = True
            ret['changes']['old'] = {}
            ret['changes']['new'] = response['content']
            ret['comment'] = 'Monitor was successfully created.'
        else:
            ret = _load_result(response, ret)

    # else something else was returned
    else:
        ret = _load_result(existing, ret)

    return ret


def modify_monitor(hostname, username, password, monitor_type, name, **kwargs):
    '''
    Modify an existing monitor.  If it does exists, only
    the parameters specified will be enforced.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    monitor_type
        The type of monitor to create
    name
        The name of the monitor to create
    kwargs
        [ arg=val ] ...

        Consult F5 BIGIP user guide for specific options for each monitor type.
        Typically, tmsh arg names are used.
    '''

    ret = {'name': name, 'changes': {}, 'result': False, 'comment': ''}

    if __opts__['test']:

        params = {
            'hostname': hostname,
            'username': username,
            'password': password,
            'monitor_type': monitor_type,
            'name': name
        }

        for key, value in six.iteritems(kwargs):
            params[key] = value

        return _test_output(ret, 'modify', params)

    #is this monitor currently configured?
    existing = __salt__['bigip.list_monitor'](hostname, username, password, monitor_type, name)

    # if it exists
    if existing['code'] == 200:

        #modify the monitor
        modified = __salt__['bigip.modify_monitor'](hostname, username, password, monitor_type, name, **kwargs)

        #was the modification successful?
        if modified['code'] == 200:
            del existing['content']['selfLink']
            del modified['content']['selfLink']
            ret = _check_for_changes('Monitor', ret, existing, modified)
        else:
            ret = _load_result(modified, ret)

    # if it doesn't exist
    elif existing['code'] == 404:
        ret['comment'] = 'A Monitor with this name was not found.'
    # else something else was returned
    else:
        ret = _load_result(existing, ret)

    return ret


def delete_monitor(hostname, username, password, monitor_type, name):
    '''
    Modify an existing monitor.  If it does exists, only
    the parameters specified will be enforced.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    monitor_type
        The type of monitor to create
    name
        The name of the monitor to create
    kwargs
        [ arg=val ] ...

        Consult F5 BIGIP user guide for specific options for each monitor type.
        Typically, tmsh arg names are used.
    '''

    ret = {'name': name, 'changes': {}, 'result': False, 'comment': ''}

    if __opts__['test']:
        return _test_output(ret, 'delete', params={
            'hostname': hostname,
            'username': username,
            'password': password,
            'monitor_type': monitor_type,
            'name': name
        })

    #is this profile currently configured?
    existing = __salt__['bigip.list_monitor'](hostname, username, password, monitor_type, name)
    # if it exists by name
    if existing['code'] == 200:

        deleted = __salt__['bigip.delete_monitor'](hostname, username, password, monitor_type, name)
        # did we get rid of it?
        if deleted['code'] == 200:
            ret['result'] = True
            ret['comment'] = 'Monitor was successfully deleted.'
            ret['changes']['old'] = existing['content']
            ret['changes']['new'] = {}
        # something bad happened
        else:
            ret = _load_result(deleted, ret)

    # not found
    elif existing['code'] == 404:
        ret['result'] = True
        ret['comment'] = 'This Monitor already does not exist. No changes made.'
        ret['changes']['old'] = {}
        ret['changes']['new'] = {}
    else:
        ret = _load_result(existing, ret)

    return ret


def list_profile(hostname, username, password, profile_type, name):
    '''
    A function to list an existing profile.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    profile_type
        The type of profile to list
    name
        The name of the profile to list
    '''

    ret = {'name': name, 'changes': {}, 'result': False, 'comment': ''}

    if __opts__['test']:
        return _test_output(ret, 'list', params={
            'hostname': hostname,
            'username': username,
            'password': password,
            'profile_type': profile_type,
            'name': name
        })

    response = __salt__['bigip.list_profile'](hostname, username, password, profile_type, name)
    return _load_result(response, ret)


def create_profile(hostname, username, password, profile_type, name, **kwargs):
    r'''
    A function to connect to a bigip device and create a profile.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    profile_type
        The type of profile to create
    name
        The name of the profile to create
    kwargs
        [ arg=val ] ...

        Consult F5 BIGIP user guide for specific options for each profile type.
        Typically, tmsh arg names are used.

    Special Characters ``|``, ``,`` and ``:`` must be escaped using ``\`` when
    used within strings.

    '''

    ret = {'name': name, 'changes': {}, 'result': False, 'comment': ''}

    if __opts__['test']:
        return _test_output(ret, 'create', params={
            'hostname': hostname,
            'username': username,
            'password': password,
            'profile_type': profile_type,
            'name': name
        })

    #is this profile currently configured?
    existing = __salt__['bigip.list_profile'](hostname, username, password, profile_type, name)

    # if it exists
    if existing['code'] == 200:

        ret['result'] = True
        ret['comment'] = 'A profile by this name currently exists.  No change made.'

    # if it doesn't exist
    elif existing['code'] == 404:

        response = __salt__['bigip.create_profile'](hostname, username, password, profile_type, name, **kwargs)

        if response['code'] == 200:
            ret['result'] = True
            ret['changes']['old'] = {}
            ret['changes']['new'] = response['content']
            ret['comment'] = 'Profile was successfully created.'
        else:
            ret = _load_result(response, ret)

    # else something else was returned
    else:
        ret = _load_result(existing, ret)

    return ret


def manage_profile(hostname, username, password, profile_type, name, **kwargs):
    '''
    Create a new profile if a monitor of this type and name does not already exists.  If it does exists, only
    the parameters specified will be enforced.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    profile_type
        The type of profile to create
    name
        The name of the profile to create
    kwargs
        [ arg=val ] ...

        Consult F5 BIGIP user guide for specific options for each profile type.
        Typically, tmsh arg names are used.
    '''

    ret = {'name': name, 'changes': {}, 'result': False, 'comment': ''}

    if __opts__['test']:

        params = {
            'hostname': hostname,
            'username': username,
            'password': password,
            'profile_type': profile_type,
            'name': name
        }

        for key, value in six.iteritems(kwargs):
            params[key] = value

        return _test_output(ret, 'manage', params)

    #is this profile currently configured?
    existing = __salt__['bigip.list_profile'](hostname, username, password, profile_type, name)

    # if it exists
    if existing['code'] == 200:

        #modify the profile
        modified = __salt__['bigip.modify_profile'](hostname, username, password, profile_type, name, **kwargs)

        #was the modification successful?
        if modified['code'] == 200:
            del existing['content']['selfLink']
            del modified['content']['selfLink']
            ret = _check_for_changes('Profile', ret, existing, modified)
        else:
            ret = _load_result(modified, ret)

    # if it doesn't exist
    elif existing['code'] == 404:

        response = __salt__['bigip.create_profile'](hostname, username, password, profile_type, name, **kwargs)
        if response['code'] == 200:
            ret['result'] = True
            ret['changes']['old'] = {}
            ret['changes']['new'] = response['content']
            ret['comment'] = 'Profile was successfully created.'
        else:
            ret = _load_result(existing, ret)

    # else something else was returned
    else:
        ret = _load_result(existing, ret)

    return ret


def modify_profile(hostname, username, password, profile_type, name, **kwargs):
    '''
    Modify an existing profile.  If it does exists, only
    the parameters specified will be enforced.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    profile_type
        The type of profile to create
    name
        The name of the profile to create
    kwargs
        [ arg=val ] ...

        Consult F5 BIGIP user guide for specific options for each monitor type.
        Typically, tmsh arg names are used.
    '''

    ret = {'name': name, 'changes': {}, 'result': False, 'comment': ''}

    if __opts__['test']:

        params = {
            'hostname': hostname,
            'username': username,
            'password': password,
            'profile_type': profile_type,
            'name': name
        }

        for key, value in six.iteritems(kwargs):
            params[key] = value

        return _test_output(ret, 'modify', params)

    #is this profile currently configured?
    existing = __salt__['bigip.list_profile'](hostname, username, password, profile_type, name)

    # if it exists
    if existing['code'] == 200:

        #modify the profile
        modified = __salt__['bigip.modify_profile'](hostname, username, password, profile_type, name, **kwargs)

        #was the modification successful?
        if modified['code'] == 200:
            del existing['content']['selfLink']
            del modified['content']['selfLink']
            ret = _check_for_changes('Profile', ret, existing, modified)
        else:
            ret = _load_result(modified, ret)

    # if it doesn't exist
    elif existing['code'] == 404:
        ret['comment'] = 'A Profile with this name was not found.'
    # else something else was returned
    else:
        ret = _load_result(existing, ret)

    return ret


def delete_profile(hostname, username, password, profile_type, name):
    '''
    Modify an existing profile.  If it does exists, only
    the parameters specified will be enforced.

    hostname
        The host/address of the bigip device
    username
        The iControl REST username
    password
        The iControl REST password
    profile_type
        The type of profile to create
    name
        The name of the profile to create
    kwargs
        [ arg=val ] ...

        Consult F5 BIGIP user guide for specific options for each profile type.
        Typically, tmsh arg names are used.
    '''

    ret = {'name': name, 'changes': {}, 'result': False, 'comment': ''}

    if __opts__['test']:
        return _test_output(ret, 'delete', params={
            'hostname': hostname,
            'username': username,
            'password': password,
            'profile_type': profile_type,
            'name': name
        })

    #is this profile currently configured?
    existing = __salt__['bigip.list_profile'](hostname, username, password, profile_type, name)

    # if it exists by name
    if existing['code'] == 200:

        deleted = __salt__['bigip.delete_profile'](hostname, username, password, profile_type, name)
        # did we get rid of it?
        if deleted['code'] == 200:
            ret['result'] = True
            ret['comment'] = 'Profile was successfully deleted.'
            ret['changes']['old'] = existing['content']
            ret['changes']['new'] = {}
        # something bad happened
        else:
            ret = _load_result(deleted, ret)

    # not found
    elif existing['code'] == 404:
        ret['result'] = True
        ret['comment'] = 'This Profile already does not exist. No changes made.'
        ret['changes']['old'] = {}
        ret['changes']['new'] = {}
    else:
        ret = _load_result(existing, ret)

    return ret



# -*- coding: utf-8 -*-

# Import Python libs
from __future__ import absolute_import

# Import Salt Testing libs
from salttesting.helpers import ensure_in_syspath
ensure_in_syspath('../../')

# Import salt libs
import integration


class TestSyndic(integration.SyndicCase):
    '''
    Validate the syndic interface by testing the test module
    '''
    def test_ping(self):
        '''
        test.ping
        '''
        self.assertTrue(self.run_function('test.ping'))

    def test_fib(self):
        '''
        test.fib
        '''
        self.assertEqual(
                self.run_function(
                    'test.fib',
                    ['20'],
                    )[0],
                6765
                )


if __name__ == '__main__':
    from integration import run_tests
    run_tests(TestSyndic)



# pylint: disable=W9903
'''
This is a shim that handles checking and updating salt thin and
then invoking thin.

This is not intended to be instantiated as a module, rather it is a
helper script used by salt.client.ssh.Single.  It is here, in a
separate file, for convenience of development.
'''

from __future__ import absolute_import

import hashlib
import tarfile
import shutil
import sys
import os
import stat
import subprocess

THIN_ARCHIVE = 'salt-thin.tgz'
EXT_ARCHIVE = 'salt-ext_mods.tgz'

# Keep these in sync with salt/defaults/exitcodes.py
EX_THIN_DEPLOY = 11
EX_THIN_CHECKSUM = 12
EX_MOD_DEPLOY = 13
EX_SCP_NOT_FOUND = 14
EX_CANTCREAT = 73


class OBJ(object):
    """An empty class for holding instance attribute values."""
    pass


OPTIONS = None
ARGS = None
# The below line is where OPTIONS can be redefined with internal options
# (rather than cli arguments) when the shim is bundled by
# client.ssh.Single._cmd_str()
# pylint: disable=block-comment-should-start-with-cardinal-space
#%%OPTS


def need_deployment():
    """
    Salt thin needs to be deployed - prep the target directory and emit the
    delimeter and exit code that signals a required deployment.
    """
    if os.path.exists(OPTIONS.saltdir):
        shutil.rmtree(OPTIONS.saltdir)
    old_umask = os.umask(0o077)
    os.makedirs(OPTIONS.saltdir)
    os.umask(old_umask)
    # Verify perms on saltdir
    euid = os.geteuid()
    dstat = os.stat(OPTIONS.saltdir)
    if dstat.st_uid != euid:
        # Attack detected, try again
        need_deployment()
    if dstat.st_mode != 16832:
        # Attack detected
        need_deployment()
    # If SUDOing then also give the super user group write permissions
    sudo_gid = os.environ.get('SUDO_GID')
    if sudo_gid:
        try:
            os.chown(OPTIONS.saltdir, -1, int(sudo_gid))
            stt = os.stat(OPTIONS.saltdir)
            os.chmod(OPTIONS.saltdir, stt.st_mode | stat.S_IWGRP | stat.S_IRGRP | stat.S_IXGRP)
        except OSError:
            sys.stdout.write('\n\nUnable to set permissions on thin directory.\nIf sudo_user is set '
                    'and is not root, be certain the user is in the same group\nas the login user')
            sys.exit(1)

    # Delimiter emitted on stdout *only* to indicate shim message to master.
    sys.stdout.write("{0}\ndeploy\n".format(OPTIONS.delimiter))
    sys.exit(EX_THIN_DEPLOY)


# Adapted from salt.utils.get_hash()
def get_hash(path, form='sha1', chunk_size=4096):
    """Generate a hash digest string for a file."""
    try:
        hash_type = getattr(hashlib, form)
    except AttributeError:
        raise ValueError('Invalid hash type: {0}'.format(form))
    with open(path, 'rb') as ifile:
        hash_obj = hash_type()
        # read the file in in chunks, not the entire file
        for chunk in iter(lambda: ifile.read(chunk_size), b''):
            hash_obj.update(chunk)
        return hash_obj.hexdigest()


def unpack_thin(thin_path):
    """Unpack the Salt thin archive."""
    tfile = tarfile.TarFile.gzopen(thin_path)
    old_umask = os.umask(0o077)
    tfile.extractall(path=OPTIONS.saltdir)
    tfile.close()
    os.umask(old_umask)
    os.unlink(thin_path)


def need_ext():
    """Signal that external modules need to be deployed."""
    sys.stdout.write("{0}\next_mods\n".format(OPTIONS.delimiter))
    sys.exit(EX_MOD_DEPLOY)


def unpack_ext(ext_path):
    """Unpack the external modules."""
    modcache = os.path.join(
            OPTIONS.saltdir,
            'running_data',
            'var',
            'cache',
            'salt',
            'minion',
            'extmods')
    tfile = tarfile.TarFile.gzopen(ext_path)
    old_umask = os.umask(0o077)
    tfile.extractall(path=modcache)
    tfile.close()
    os.umask(old_umask)
    os.unlink(ext_path)
    ver_path = os.path.join(modcache, 'ext_version')
    ver_dst = os.path.join(OPTIONS.saltdir, 'ext_version')
    shutil.move(ver_path, ver_dst)


def main(argv):  # pylint: disable=W0613
    """Main program body"""
    thin_path = os.path.join(OPTIONS.saltdir, THIN_ARCHIVE)
    if os.path.isfile(thin_path):
        if OPTIONS.checksum != get_hash(thin_path, OPTIONS.hashfunc):
            need_deployment()
        unpack_thin(thin_path)
        # Salt thin now is available to use
    else:
        scpstat = subprocess.Popen(['/bin/sh', '-c', 'command -v scp']).wait()
        if scpstat != 0:
            sys.exit(EX_SCP_NOT_FOUND)

        if not os.path.exists(OPTIONS.saltdir):
            need_deployment()

        if not os.path.isdir(OPTIONS.saltdir):
            sys.stderr.write(
                'ERROR: salt path "{0}" exists but is'
                ' not a directory\n'.format(OPTIONS.saltdir)
            )
            sys.exit(EX_CANTCREAT)

        version_path = os.path.join(OPTIONS.saltdir, 'version')
        if not os.path.exists(version_path) or not os.path.isfile(version_path):
            sys.stderr.write(
                'WARNING: Unable to locate current thin '
                ' version: {0}.\n'.format(version_path)
            )
            need_deployment()
        with open(version_path, 'r') as vpo:
            cur_version = vpo.readline().strip()
        if cur_version != OPTIONS.version:
            sys.stderr.write(
                'WARNING: current thin version {0}'
                ' is not up-to-date with {1}.\n'.format(
                    cur_version, OPTIONS.version
                )
            )
            need_deployment()
        # Salt thin exists and is up-to-date - fall through and use it

    salt_call_path = os.path.join(OPTIONS.saltdir, 'salt-call')
    if not os.path.isfile(salt_call_path):
        sys.stderr.write('ERROR: thin is missing "{0}"\n'.format(salt_call_path))
        need_deployment()

    with open(os.path.join(OPTIONS.saltdir, 'minion'), 'w') as config:
        config.write(OPTIONS.config + '\n')
    if OPTIONS.ext_mods:
        ext_path = os.path.join(OPTIONS.saltdir, EXT_ARCHIVE)
        if os.path.exists(ext_path):
            unpack_ext(ext_path)
        else:
            version_path = os.path.join(OPTIONS.saltdir, 'ext_version')
            if not os.path.exists(version_path) or not os.path.isfile(version_path):
                need_ext()
            with open(version_path, 'r') as vpo:
                cur_version = vpo.readline().strip()
            if cur_version != OPTIONS.ext_mods:
                need_ext()
    # Fix parameter passing issue
    if len(ARGS) == 1:
        argv_prepared = ARGS[0].split()
    else:
        argv_prepared = ARGS

    salt_argv = [
        sys.executable,
        salt_call_path,
        '--retcode-passthrough',
        '--local',
        '--metadata',
        '--out', 'json',
        '-l', 'quiet',
        '-c', OPTIONS.saltdir,
        '--',
    ] + argv_prepared

    sys.stderr.write('SALT_ARGV: {0}\n'.format(salt_argv))

    # Only emit the delimiter on *both* stdout and stderr when completely successful.
    # Yes, the flush() is necessary.
    sys.stdout.write(OPTIONS.delimiter + '\n')
    sys.stdout.flush()
    if not OPTIONS.tty:
        sys.stderr.write(OPTIONS.delimiter + '\n')
        sys.stderr.flush()
    if OPTIONS.cmd_umask is not None:
        old_umask = os.umask(OPTIONS.cmd_umask)
    if OPTIONS.tty:
        stdout, _ = subprocess.Popen(salt_argv, stdout=subprocess.PIPE, stderr=subprocess.PIPE).communicate()
        sys.stdout.write(stdout)
        sys.stdout.flush()
        if OPTIONS.wipe:
            shutil.rmtree(OPTIONS.saltdir)
    elif OPTIONS.wipe:
        subprocess.call(salt_argv)
        shutil.rmtree(OPTIONS.saltdir)
    else:
        subprocess.call(salt_argv)
    if OPTIONS.cmd_umask is not None:
        os.umask(old_umask)

if __name__ == '__main__':
    sys.exit(main(sys.argv))



# -*- coding: utf-8 -*-
'''
integration tests for mac_system
'''

# Import python libs
from __future__ import absolute_import, print_function
import random
import string

# Import Salt Testing libs
from salttesting import skipIf
from salttesting.helpers import ensure_in_syspath, destructiveTest
from salt.ext.six.moves import range
ensure_in_syspath('../../')

# Import salt libs
import integration
import salt.utils


def __random_string(size=6):
    '''
    Generates a random username
    '''
    return 'RS-' + ''.join(
        random.choice(string.ascii_uppercase + string.digits)
        for x in range(size)
    )


SET_COMPUTER_NAME = __random_string()
SET_SUBNET_NAME = __random_string()


@skipIf(not salt.utils.is_darwin()
        or not salt.utils.which('systemsetup')
        or salt.utils.get_uid(salt.utils.get_user()) != 0, 'Test requirements not met')
class MacSystemModuleTest(integration.ModuleCase):
    '''
    Validate the mac_system module
    '''
    ATRUN_ENABLED = False
    REMOTE_LOGIN_ENABLED = False
    REMOTE_EVENTS_ENABLED = False
    COMPUTER_NAME = ''
    SUBNET_NAME = ''
    KEYBOARD_DISABLED = False

    def setUp(self):
        '''
        Get current settings
        '''
        self.ATRUN_ENABLED = self.run_function('service.enabled', ['com.apple.atrun'])
        self.REMOTE_LOGIN_ENABLED = self.run_function('system.get_remote_login')
        self.REMOTE_EVENTS_ENABLED = self.run_function('system.get_remote_events')
        self.COMPUTER_NAME = self.run_function('system.get_computer_name')
        self.SUBNET_NAME = self.run_function('system.get_subnet_name')
        self.KEYBOARD_DISABLED = self.run_function('system.get_disable_keyboard_on_lock')

    def tearDown(self):
        '''
        Reset to original settings
        '''
        if not self.ATRUN_ENABLED:
            atrun = '/System/Library/LaunchDaemons/com.apple.atrun.plist'
            self.run_function('service.stop', [atrun])

        self.run_function('system.set_remote_login', [self.REMOTE_LOGIN_ENABLED])
        self.run_function('system.set_remote_events', [self.REMOTE_EVENTS_ENABLED])
        self.run_function('system.set_computer_name', [self.COMPUTER_NAME])
        self.run_function('system.set_subnet_name', [self.SUBNET_NAME])
        self.run_function('system.set_disable_keyboard_on_lock',
                          [self.KEYBOARD_DISABLED])

    @destructiveTest
    def test_get_set_remote_login(self):
        '''
        Test system.get_remote_login
        Test system.set_remote_login
        '''
        # Normal Functionality
        self.assertTrue(self.run_function('system.set_remote_login', [True]))
        self.assertTrue(self.run_function('system.get_remote_login'))
        self.assertTrue(self.run_function('system.set_remote_login', [False]))
        self.assertFalse(self.run_function('system.get_remote_login'))

        # Test valid input
        self.assertTrue(self.run_function('system.set_remote_login', [True]))
        self.assertTrue(self.run_function('system.set_remote_login', [False]))
        self.assertTrue(self.run_function('system.set_remote_login', ['yes']))
        self.assertTrue(self.run_function('system.set_remote_login', ['no']))
        self.assertTrue(self.run_function('system.set_remote_login', ['On']))
        self.assertTrue(self.run_function('system.set_remote_login', ['Off']))
        self.assertTrue(self.run_function('system.set_remote_login', [1]))
        self.assertTrue(self.run_function('system.set_remote_login', [0]))

        # Test invalid input
        self.assertIn(
            'Invalid String Value for Enabled',
            self.run_function('system.set_remote_login', ['spongebob']))

    @destructiveTest
    def test_get_set_remote_events(self):
        '''
        Test system.get_remote_events
        Test system.set_remote_events
        '''
        # Normal Functionality
        self.assertTrue(self.run_function('system.set_remote_events', [True]))
        self.assertTrue(self.run_function('system.get_remote_events'))
        self.assertTrue(self.run_function('system.set_remote_events', [False]))
        self.assertFalse(self.run_function('system.get_remote_events'))

        # Test valid input
        self.assertTrue(self.run_function('system.set_remote_events', [True]))
        self.assertTrue(self.run_function('system.set_remote_events', [False]))
        self.assertTrue(self.run_function('system.set_remote_events', ['yes']))
        self.assertTrue(self.run_function('system.set_remote_events', ['no']))
        self.assertTrue(self.run_function('system.set_remote_events', ['On']))
        self.assertTrue(self.run_function('system.set_remote_events', ['Off']))
        self.assertTrue(self.run_function('system.set_remote_events', [1]))
        self.assertTrue(self.run_function('system.set_remote_events', [0]))

        # Test invalid input
        self.assertIn(
            'Invalid String Value for Enabled',
            self.run_function('system.set_remote_events', ['spongebob']))

    @destructiveTest
    def test_get_set_computer_name(self):
        '''
        Test system.get_computer_name
        Test system.set_computer_name
        '''
        self.assertTrue(
            self.run_function('system.set_computer_name', [SET_COMPUTER_NAME]))
        self.assertEqual(
            self.run_function('system.get_computer_name'),
            SET_COMPUTER_NAME)

    @destructiveTest
    def test_get_set_subnet_name(self):
        '''
        Test system.get_subnet_name
        Test system.set_subnet_name
        '''
        self.assertTrue(
            self.run_function('system.set_subnet_name', [SET_SUBNET_NAME]))
        self.assertEqual(
            self.run_function('system.get_subnet_name'),
            SET_SUBNET_NAME)

    def test_get_list_startup_disk(self):
        '''
        Test system.get_startup_disk
        Test system.list_startup_disks
        Don't know how to test system.set_startup_disk as there's usually only
        one startup disk available on a system
        '''
        # Test list and get
        ret = self.run_function('system.list_startup_disks')
        self.assertIsInstance(ret, list)
        self.assertIn(self.run_function('system.get_startup_disk'), ret)

        # Test passing set a bad disk
        self.assertIn(
            'Invalid value passed for path.',
            self.run_function('system.set_startup_disk', ['spongebob']))

    @skipIf(True, 'Skip this test until mac fixes it.')
    def test_get_set_restart_delay(self):
        '''
        Test system.get_restart_delay
        Test system.set_restart_delay
        system.set_restart_delay does not work due to an apple bug, see docs
        may need to disable this test as we can't control the delay value
        '''
        # Normal Functionality
        self.assertTrue(self.run_function('system.set_restart_delay', [90]))
        self.assertEqual(
            self.run_function('system.get_restart_delay'),
            '90 seconds')

        # Pass set bad value for seconds
        self.assertIn(
            'Invalid value passed for seconds.',
            self.run_function('system.set_restart_delay', [70]))

    def test_get_set_disable_keyboard_on_lock(self):
        '''
        Test system.get_disable_keyboard_on_lock
        Test system.set_disable_keyboard_on_lock
        '''
        # Normal Functionality
        self.assertTrue(
            self.run_function('system.set_disable_keyboard_on_lock', [True]))
        self.assertTrue(
            self.run_function('system.get_disable_keyboard_on_lock'))

        self.assertTrue(
            self.run_function('system.set_disable_keyboard_on_lock', [False]))
        self.assertFalse(
            self.run_function('system.get_disable_keyboard_on_lock'))

        # Test valid input
        self.assertTrue(
            self.run_function('system.set_disable_keyboard_on_lock', [True]))
        self.assertTrue(
            self.run_function('system.set_disable_keyboard_on_lock', [False]))
        self.assertTrue(
            self.run_function('system.set_disable_keyboard_on_lock', ['yes']))
        self.assertTrue(
            self.run_function('system.set_disable_keyboard_on_lock', ['no']))
        self.assertTrue(
            self.run_function('system.set_disable_keyboard_on_lock', ['On']))
        self.assertTrue(
            self.run_function('system.set_disable_keyboard_on_lock', ['Off']))
        self.assertTrue(
            self.run_function('system.set_disable_keyboard_on_lock', [1]))
        self.assertTrue(
            self.run_function('system.set_disable_keyboard_on_lock', [0]))

        # Test invalid input
        self.assertIn(
            'Invalid String Value for Enabled',
            self.run_function('system.set_disable_keyboard_on_lock',
                              ['spongebob']))

    @skipIf(True, 'Skip this test until mac fixes it.')
    def test_get_set_boot_arch(self):
        '''
        Test system.get_boot_arch
        Test system.set_boot_arch
        system.set_boot_arch does not work due to an apple bug, see docs
        may need to disable this test as we can't set the boot architecture
        '''
        # Normal Functionality
        self.assertTrue(self.run_function('system.set_boot_arch', ['i386']))
        self.assertEqual(self.run_function('system.get_boot_arch'), 'i386')
        self.assertTrue(self.run_function('system.set_boot_arch', ['default']))
        self.assertEqual(self.run_function('system.get_boot_arch'), 'default')

        # Test invalid input
        self.assertIn(
            'Invalid value passed for arch',
            self.run_function('system.set_boot_arch', ['spongebob']))


if __name__ == '__main__':
    from integration import run_tests
    run_tests(MacSystemModuleTest)



# -*- coding: utf-8 -*-
'''
Integration tests for the mac_desktop execution module.
'''

# Import Python Libs
from __future__ import absolute_import
import os

# Import Salt Testing Libs
from salttesting import skipIf
from salttesting.helpers import (
    destructiveTest,
    ensure_in_syspath,
    requires_system_grains
)

ensure_in_syspath('../../')

# Import Salt Libs
import integration


@skipIf(os.geteuid() != 0, 'You must be logged in as root to run this test')
class MacDesktopTestCase(integration.ModuleCase):
    '''
    Integration tests for the mac_desktop module.
    '''

    def setUp(self):
        '''
        Sets up test requirements.
        '''
        os_grain = self.run_function('grains.item', ['kernel'])
        if os_grain['kernel'] not in 'Darwin':
            self.skipTest(
                'Test not applicable to \'{kernel}\' kernel'.format(
                    **os_grain
                )
            )

    @requires_system_grains
    def test_get_output_volume(self, grains=None):
        '''
        Tests the return of get_output_volume.
        '''
        ret = self.run_function('desktop.get_output_volume')
        self.assertIsNotNone(ret)

    @destructiveTest
    @requires_system_grains
    def test_set_output_volume(self, grains=None):
        '''
        Tests the return of set_output_volume.
        '''
        current_vol = self.run_function('desktop.get_output_volume')
        to_set = 10
        if current_vol == str(to_set):
            to_set += 2
        new_vol = self.run_function('desktop.set_output_volume', [str(to_set)])
        check_vol = self.run_function('desktop.get_output_volume')
        self.assertEqual(new_vol, check_vol)

        # Set volume back to what it was before
        self.run_function('desktop.set_output_volume', [current_vol])

    @destructiveTest
    @requires_system_grains
    def test_screensaver(self, grains=None):
        '''
        Tests the return of the screensaver function.
        '''
        self.assertTrue(
            self.run_function('desktop.screensaver')
        )

    @destructiveTest
    @requires_system_grains
    def test_lock(self, grains=None):
        '''
        Tests the return of the lock function.
        '''
        self.assertTrue(
            self.run_function('desktop.lock')
        )

    @destructiveTest
    @requires_system_grains
    def test_say(self, grains=None):
        '''
        Tests the return of the say function.
        '''
        self.assertTrue(
            self.run_function('desktop.say', ['hello', 'world'])
        )


if __name__ == '__main__':
    from integration import run_tests
    run_tests(MacDesktopTestCase)



# -*- coding: utf-8 -*-
'''
    :codeauthor: :email:`Jayesh Kariya <jayeshk@saltstack.com>`
'''

# Import Python Libs
from __future__ import absolute_import

# Import Salt Testing Libs
from salttesting import TestCase, skipIf
from salttesting.mock import (
    MagicMock,
    patch,
    NO_MOCK,
    NO_MOCK_REASON
)
from salttesting.helpers import ensure_in_syspath

ensure_in_syspath('../../')

# Import Salt Libs
from salt.utils import etcd_util
try:
    from urllib3.exceptions import ReadTimeoutError, MaxRetryError
    HAS_URLLIB3 = True
except ImportError:
    HAS_URLLIB3 = False

try:
    import etcd
    HAS_ETCD = True
except ImportError:
    HAS_ETCD = False


@skipIf(HAS_URLLIB3 is False, 'urllib3 module must be installed.')
@skipIf(HAS_ETCD is False, 'python-etcd module must be installed.')
@skipIf(NO_MOCK, NO_MOCK_REASON)
class EtcdUtilTestCase(TestCase):
    '''
    Test cases for salt.utils.etcd_util
    '''
    # 'get_' function tests: 1

    @patch('etcd.Client', autospec=True)
    def test_read(self, mock):
        '''
        Test to make sure we interact with etcd correctly
        '''

        etcd_client = mock.return_value
        etcd_return = MagicMock(value='salt')
        etcd_client.read.return_value = etcd_return
        client = etcd_util.EtcdClient({})

        self.assertEqual(client.read('/salt'), etcd_return)
        etcd_client.read.assert_called_with('/salt', recursive=False, wait=False, timeout=None)

        client.read('salt', True, True, 10, 5)
        etcd_client.read.assert_called_with('salt', recursive=True, wait=True, timeout=10, waitIndex=5)

        etcd_client.read.side_effect = etcd.EtcdKeyNotFound
        self.assertRaises(etcd.EtcdKeyNotFound, client.read, 'salt')

        etcd_client.read.side_effect = etcd.EtcdConnectionFailed
        self.assertRaises(etcd.EtcdConnectionFailed, client.read, 'salt')

        etcd_client.read.side_effect = etcd.EtcdValueError
        self.assertRaises(etcd.EtcdValueError, client.read, 'salt')

        etcd_client.read.side_effect = ValueError
        self.assertRaises(ValueError, client.read, 'salt')

        etcd_client.read.side_effect = ReadTimeoutError(None, None, None)
        self.assertRaises(etcd.EtcdConnectionFailed, client.read, 'salt')

        etcd_client.read.side_effect = MaxRetryError(None, None)
        self.assertRaises(etcd.EtcdConnectionFailed, client.read, 'salt')

    @patch('etcd.Client')
    def test_get(self, mock):
        '''
        Test if it get a value from etcd, by direct path
        '''
        client = etcd_util.EtcdClient({})

        with patch.object(client, 'read', autospec=True) as mock:
            mock.return_value = MagicMock(value='stack')
            self.assertEqual(client.get('salt'), 'stack')
            mock.assert_called_with('salt', recursive=False)

            self.assertEqual(client.get('salt', recurse=True), 'stack')
            mock.assert_called_with('salt', recursive=True)

            mock.side_effect = etcd.EtcdKeyNotFound()
            self.assertEqual(client.get('not-found'), None)

            mock.side_effect = etcd.EtcdConnectionFailed()
            self.assertEqual(client.get('watching'), None)

            # python 2.6 test
            mock.side_effect = ValueError
            self.assertEqual(client.get('not-found'), None)

            mock.side_effect = Exception
            self.assertRaises(Exception, client.get, 'some-error')

    @patch('etcd.Client')
    def test_tree(self, mock):
        '''
        Test recursive gets
        '''
        client = etcd_util.EtcdClient({})

        with patch.object(client, 'read', autospec=True) as mock:

            c1, c2 = MagicMock(), MagicMock()
            c1.__iter__.return_value = [
                MagicMock(key='/x/a', value='1'),
                MagicMock(key='/x/b', value='2'),
                MagicMock(key='/x/c', dir=True)]
            c2.__iter__.return_value = [
                MagicMock(key='/x/c/d', value='3')
            ]
            mock.side_effect = iter([
                MagicMock(children=c1),
                MagicMock(children=c2)
            ])
            self.assertDictEqual(client.tree('/x'), {'a': '1', 'b': '2', 'c': {'d': '3'}})
            mock.assert_any_call('/x')
            mock.assert_any_call('/x/c')

            mock.side_effect = etcd.EtcdKeyNotFound()
            self.assertEqual(client.tree('not-found'), None)

            mock.side_effect = ValueError
            self.assertEqual(client.tree('/x'), None)

            mock.side_effect = Exception
            self.assertRaises(Exception, client.tree, 'some-error')

    @patch('etcd.Client')
    def test_ls(self, mock):
        client = etcd_util.EtcdClient({})

        with patch.object(client, 'read', autospec=True) as mock:
            c1 = MagicMock()
            c1.__iter__.return_value = [
                MagicMock(key='/x/a', value='1'),
                MagicMock(key='/x/b', value='2'),
                MagicMock(key='/x/c', dir=True)]
            mock.return_value = MagicMock(children=c1)
            self.assertEqual(client.ls('/x'), {'/x': {'/x/a': '1', '/x/b': '2', '/x/c/': {}}})
            mock.assert_called_with('/x')

            mock.side_effect = etcd.EtcdKeyNotFound()
            self.assertEqual(client.ls('/not-found'), {})

            mock.side_effect = Exception
            self.assertRaises(Exception, client.tree, 'some-error')

    @patch('etcd.Client', autospec=True)
    def test_write(self, mock):
        client = etcd_util.EtcdClient({})
        etcd_client = mock.return_value

        etcd_client.write.return_value = MagicMock(value='salt')
        self.assertEqual(client.write('/some-key', 'salt'), 'salt')
        etcd_client.write.assert_called_with('/some-key', 'salt', ttl=None, dir=False)

        self.assertEqual(client.write('/some-key', 'salt', ttl=5), 'salt')
        etcd_client.write.assert_called_with('/some-key', 'salt', ttl=5, dir=False)

        etcd_client.write.return_value = MagicMock(dir=True)
        self.assertEqual(client.write('/some-dir', 'salt', ttl=0, directory=True), True)
        etcd_client.write.assert_called_with('/some-dir', None, ttl=0, dir=True)

        etcd_client.write.side_effect = etcd.EtcdRootReadOnly()
        self.assertEqual(client.write('/', 'some-val'), None)

        etcd_client.write.side_effect = etcd.EtcdNotFile()
        self.assertEqual(client.write('/some-key', 'some-val'), None)

        etcd_client.write.side_effect = etcd.EtcdNotDir()
        self.assertEqual(client.write('/some-dir', 'some-val'), None)

        etcd_client.write.side_effect = MaxRetryError(None, None)
        self.assertEqual(client.write('/some-key', 'some-val'), None)

        etcd_client.write.side_effect = ValueError
        self.assertEqual(client.write('/some-key', 'some-val'), None)

        etcd_client.write.side_effect = Exception
        self.assertRaises(Exception, client.set, 'some-key', 'some-val')

    @patch('etcd.Client', autospec=True)
    def test_flatten(self, mock):
        client = etcd_util.EtcdClient({})
        some_data = {
            '/x/y/a': '1',
            'x': {
                'y': {
                    'b': '2'
                }
            },
            'm/j/': '3',
            'z': '4',
            'd': {},
        }

        result_path = {
            '/test/x/y/a': '1',
            '/test/x/y/b': '2',
            '/test/m/j': '3',
            '/test/z': '4',
            '/test/d': {},
        }

        result_nopath = {
            '/x/y/a': '1',
            '/x/y/b': '2',
            '/m/j': '3',
            '/z': '4',
            '/d': {},
        }

        result_root = {
            '/x/y/a': '1',
            '/x/y/b': '2',
            '/m/j': '3',
            '/z': '4',
            '/d': {},
        }

        self.assertEqual(client._flatten(some_data, path='/test'), result_path)
        self.assertEqual(client._flatten(some_data, path='/'), result_root)
        self.assertEqual(client._flatten(some_data), result_nopath)

    @patch('etcd.Client', autospec=True)
    def test_update(self, mock):
        client = etcd_util.EtcdClient({})
        some_data = {
            '/x/y/a': '1',
            'x': {
                'y': {
                    'b': '3'
                }
            },
            'm/j/': '3',
            'z': '4',
            'd': {},
        }

        result = {
            '/test/x/y/a': '1',
            '/test/x/y/b': '2',
            '/test/m/j': '3',
            '/test/z': '4',
            '/test/d': True,
        }

        flatten_result = {
            '/test/x/y/a': '1',
            '/test/x/y/b': '2',
            '/test/m/j': '3',
            '/test/z': '4',
            '/test/d': {}
        }
        client._flatten = MagicMock(return_value=flatten_result)

        self.assertEqual(client.update('/some/key', path='/blah'), None)

        with patch.object(client, 'write', autospec=True) as write_mock:
            def write_return(key, val, ttl=None, directory=None):
                return result.get(key, None)
            write_mock.side_effect = write_return
            self.assertDictEqual(client.update(some_data, path='/test'), result)
            client._flatten.assert_called_with(some_data, '/test')
            self.assertEqual(write_mock.call_count, 5)

    @patch('etcd.Client', autospec=True)
    def test_rm(self, mock):
        etcd_client = mock.return_value
        client = etcd_util.EtcdClient({})

        etcd_client.delete.return_value = True
        self.assertEqual(client.rm('/some-key'), True)
        etcd_client.delete.assert_called_with('/some-key', recursive=False)
        self.assertEqual(client.rm('/some-dir', recurse=True), True)
        etcd_client.delete.assert_called_with('/some-dir', recursive=True)

        etcd_client.delete.side_effect = etcd.EtcdNotFile()
        self.assertEqual(client.rm('/some-dir'), None)

        etcd_client.delete.side_effect = etcd.EtcdDirNotEmpty()
        self.assertEqual(client.rm('/some-key'), None)

        etcd_client.delete.side_effect = etcd.EtcdRootReadOnly()
        self.assertEqual(client.rm('/'), None)

        etcd_client.delete.side_effect = ValueError
        self.assertEqual(client.rm('/some-dir'), None)

        etcd_client.delete.side_effect = Exception
        self.assertRaises(Exception, client.rm, 'some-dir')

    @patch('etcd.Client', autospec=True)
    def test_watch(self, client_mock):
        client = etcd_util.EtcdClient({})

        with patch.object(client, 'read', autospec=True) as mock:
            mock.return_value = MagicMock(value='stack', key='/some-key', modifiedIndex=1, dir=False)
            self.assertDictEqual(client.watch('/some-key'),
                                 {'value': 'stack', 'key': '/some-key', 'mIndex': 1, 'changed': True, 'dir': False})
            mock.assert_called_with('/some-key', wait=True, recursive=False, timeout=0, waitIndex=None)

            mock.side_effect = iter([etcd_util.EtcdUtilWatchTimeout, mock.return_value])
            self.assertDictEqual(client.watch('/some-key'),
                                 {'value': 'stack', 'changed': False, 'mIndex': 1, 'key': '/some-key', 'dir': False})

            mock.side_effect = iter([etcd_util.EtcdUtilWatchTimeout, etcd.EtcdKeyNotFound])
            self.assertEqual(client.watch('/some-key'),
                             {'value': None, 'changed': False, 'mIndex': 0, 'key': '/some-key', 'dir': False})

            mock.side_effect = iter([etcd_util.EtcdUtilWatchTimeout, ValueError])
            self.assertEqual(client.watch('/some-key'), {})

            mock.side_effect = None
            mock.return_value = MagicMock(value='stack', key='/some-key', modifiedIndex=1, dir=True)
            self.assertDictEqual(client.watch('/some-dir', recurse=True, timeout=5, index=10),
                                 {'value': 'stack', 'key': '/some-key', 'mIndex': 1, 'changed': True, 'dir': True})
            mock.assert_called_with('/some-dir', wait=True, recursive=True, timeout=5, waitIndex=10)

            mock.side_effect = MaxRetryError(None, None)
            self.assertEqual(client.watch('/some-key'), {})

            mock.side_effect = etcd.EtcdConnectionFailed()
            self.assertEqual(client.watch('/some-key'), {})

            mock.return_value = None
            self.assertEqual(client.watch('/some-key'), {})


if __name__ == '__main__':
    from integration import run_tests
    run_tests(EtcdUtilTestCase, needs_daemon=False)



# -*- coding: utf-8 -*-
'''
    :codeauthor: :email:`Jayesh Kariya <jayeshk@saltstack.com>`
'''
# Import Python libs
from __future__ import absolute_import

# Import Salt Testing Libs
from salttesting import skipIf, TestCase
from salttesting.mock import (
    NO_MOCK,
    NO_MOCK_REASON,
    MagicMock,
    patch
)

from salttesting.helpers import ensure_in_syspath

ensure_in_syspath('../../')

# Import Salt Libs
from salt.states import process

process.__opts__ = {}
process.__salt__ = {}


@skipIf(NO_MOCK, NO_MOCK_REASON)
class ProcessTestCase(TestCase):
    '''
    Test cases for salt.states.process
    '''
    # 'absent' function tests: 1

    def test_absent(self):
        '''
        Test to ensures that the named command is not running.
        '''
        name = 'apache2'

        ret = {'name': name,
               'changes': {},
               'result': None,
               'comment': ''}

        mock = MagicMock(return_value='')
        with patch.dict(process.__salt__, {'ps.pgrep': mock,
                                           'ps.pkill': mock}):
            with patch.dict(process.__opts__, {'test': True}):
                comt = ('No matching processes running')
                ret.update({'comment': comt})
                self.assertDictEqual(process.absent(name), ret)

            with patch.dict(process.__opts__, {'test': False}):
                ret.update({'result': True})
                self.assertDictEqual(process.absent(name), ret)


if __name__ == '__main__':
    from integration import run_tests
    run_tests(ProcessTestCase, needs_daemon=False)



# -*- coding: utf-8 -*-
'''
Generate the salt thin tarball from the installed python files
'''

# Import python libs
from __future__ import absolute_import

import os
import sys
import json
import shutil
import tarfile
import zipfile
import tempfile
import subprocess

# Import third party libs
import jinja2
import yaml
import msgpack
import salt.ext.six as _six
import tornado

# pylint: disable=import-error,no-name-in-module
try:
    import certifi
    HAS_CERTIFI = True
except ImportError:
    HAS_CERTIFI = False

try:
    import singledispatch
    HAS_SINGLEDISPATCH = True
except ImportError:
    HAS_SINGLEDISPATCH = False

try:
    import singledispatch_helpers
    HAS_SINGLEDISPATCH_HELPERS = True
except ImportError:
    HAS_SINGLEDISPATCH_HELPERS = False

try:
    import backports_abc
    HAS_BACKPORTS_ABC = True
except ImportError:
    HAS_BACKPORTS_ABC = False

try:
    import markupsafe
    HAS_MARKUPSAFE = True
except ImportError:
    # Older jinja does not need markupsafe
    HAS_MARKUPSAFE = False

# pylint: enable=import-error,no-name-in-module

try:
    # Older python where the backport from pypi is installed
    from backports import ssl_match_hostname
    HAS_SSL_MATCH_HOSTNAME = True
except ImportError:
    # Other older python we use our bundled copy
    try:
        from salt.ext import ssl_match_hostname
        HAS_SSL_MATCH_HOSTNAME = True
    except ImportError:
        HAS_SSL_MATCH_HOSTNAME = False

# Import salt libs
import salt
import salt.utils
import salt.exceptions

SALTCALL = '''
import os
import sys

sys.path.insert(
    0,
    os.path.join(
        os.path.dirname(__file__),
        'py{0[0]}'.format(sys.version_info)
    )
)

from salt.scripts import salt_call
if __name__ == '__main__':
    salt_call()
'''


def thin_path(cachedir):
    '''
    Return the path to the thin tarball
    '''
    return os.path.join(cachedir, 'thin', 'thin.tgz')


def get_tops(extra_mods='', so_mods=''):
    tops = [
            os.path.dirname(salt.__file__),
            os.path.dirname(jinja2.__file__),
            os.path.dirname(yaml.__file__),
            os.path.dirname(tornado.__file__),
            os.path.dirname(msgpack.__file__),
            ]

    tops.append(_six.__file__.replace('.pyc', '.py'))

    if HAS_CERTIFI:
        tops.append(os.path.dirname(certifi.__file__))

    if HAS_SINGLEDISPATCH:
        tops.append(singledispatch.__file__.replace('.pyc', '.py'))

    if HAS_SINGLEDISPATCH_HELPERS:
        tops.append(singledispatch_helpers.__file__.replace('.pyc', '.py'))

    if HAS_BACKPORTS_ABC:
        tops.append(backports_abc.__file__.replace('.pyc', '.py'))

    if HAS_SSL_MATCH_HOSTNAME:
        tops.append(os.path.dirname(os.path.dirname(ssl_match_hostname.__file__)))

    for mod in [m for m in extra_mods.split(',') if m]:
        if mod not in locals() and mod not in globals():
            try:
                locals()[mod] = __import__(mod)
                moddir, modname = os.path.split(locals()[mod].__file__)
                base, ext = os.path.splitext(modname)
                if base == '__init__':
                    tops.append(moddir)
                else:
                    tops.append(os.path.join(moddir, base + '.py'))
            except ImportError:
                # Not entirely sure this is the right thing, but the only
                # options seem to be 1) fail, 2) spew errors, or 3) pass.
                # Nothing else in here spits errors, and the markupsafe code
                # doesn't bail on import failure, so I followed that lead.
                # And of course, any other failure still S/T's.
                pass
    for mod in [m for m in so_mods.split(',') if m]:
        try:
            locals()[mod] = __import__(mod)
            tops.append(locals()[mod].__file__)
        except ImportError:
            pass   # As per comment above
    if HAS_MARKUPSAFE:
        tops.append(os.path.dirname(markupsafe.__file__))

    return tops


def gen_thin(cachedir, extra_mods='', overwrite=False, so_mods='',
             python2_bin='python2', python3_bin='python3'):
    '''
    Generate the salt-thin tarball and print the location of the tarball
    Optional additional mods to include (e.g. mako) can be supplied as a comma
    delimited string.  Permits forcing an overwrite of the output file as well.

    CLI Example:

    .. code-block:: bash

        salt-run thin.generate
        salt-run thin.generate mako
        salt-run thin.generate mako,wempy 1
        salt-run thin.generate overwrite=1
    '''
    thindir = os.path.join(cachedir, 'thin')
    if not os.path.isdir(thindir):
        os.makedirs(thindir)
    thintar = os.path.join(thindir, 'thin.tgz')
    thinver = os.path.join(thindir, 'version')
    pythinver = os.path.join(thindir, '.thin-gen-py-version')
    salt_call = os.path.join(thindir, 'salt-call')
    with salt.utils.fopen(salt_call, 'w+') as fp_:
        fp_.write(SALTCALL)
    if os.path.isfile(thintar):
        if not overwrite:
            if os.path.isfile(thinver):
                with salt.utils.fopen(thinver) as fh_:
                    overwrite = fh_.read() != salt.version.__version__
                if overwrite is False and os.path.isfile(pythinver):
                    with salt.utils.fopen(pythinver) as fh_:
                        overwrite = fh_.read() != str(sys.version_info[0])
            else:
                overwrite = True

        if overwrite:
            try:
                os.remove(thintar)
            except OSError:
                pass
        else:
            return thintar
    if _six.PY3:
        # Let's check for the minimum python 2 version requirement, 2.6
        py_shell_cmd = (
            python2_bin + ' -c \'from __future__ import print_function; import sys; '
            'print("{0}.{1}".format(*(sys.version_info[:2])));\''
        )
        cmd = subprocess.Popen(py_shell_cmd, stdout=subprocess.PIPE, shell=True)
        stdout, _ = cmd.communicate()
        if cmd.returncode == 0:
            py2_version = tuple(int(n) for n in stdout.decode('utf-8').strip().split('.'))
            if py2_version < (2, 6):
                # Bail!
                raise salt.exceptions.SaltSystemExit(
                    'The minimum required python version to run salt-ssh is "2.6".'
                    'The version reported by "{0}" is "{1}". Please try "salt-ssh '
                    '--python2-bin=<path-to-python-2.6-binary-or-higher>".'.format(python2_bin,
                                                                                stdout.strip())
                )
    elif sys.version_info < (2, 6):
        # Bail! Though, how did we reached this far in the first place.
        raise salt.exceptions.SaltSystemExit(
            'The minimum required python version to run salt-ssh is "2.6".'
        )

    tops_py_version_mapping = {}
    tops = get_tops(extra_mods=extra_mods, so_mods=so_mods)
    if _six.PY2:
        tops_py_version_mapping['2'] = tops
    else:
        tops_py_version_mapping['3'] = tops

    # TODO: Consider putting known py2 and py3 compatible libs in it's own sharable directory.
    #       This would reduce the thin size.
    if _six.PY2 and sys.version_info[0] == 2:
        # Get python 3 tops
        py_shell_cmd = (
            python3_bin + ' -c \'import sys; import json; import salt.utils.thin; '
            'print(json.dumps(salt.utils.thin.get_tops(**(json.loads(sys.argv[1]))))); exit(0);\' '
            '\'{0}\''.format(json.dumps({'extra_mods': extra_mods, 'so_mods': so_mods}))
        )
        cmd = subprocess.Popen(py_shell_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        stdout, stderr = cmd.communicate()
        if cmd.returncode == 0:
            try:
                tops = json.loads(stdout)
                tops_py_version_mapping['3'] = tops
            except ValueError:
                pass
    if _six.PY3 and sys.version_info[0] == 3:
        # Get python 2 tops
        py_shell_cmd = (
            python2_bin + ' -c \'from __future__ import print_function; '
            'import sys; import json; import salt.utils.thin; '
            'print(json.dumps(salt.utils.thin.get_tops(**(json.loads(sys.argv[1]))))); exit(0);\' '
            '\'{0}\''.format(json.dumps({'extra_mods': extra_mods, 'so_mods': so_mods}))
        )
        cmd = subprocess.Popen(py_shell_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        stdout, stderr = cmd.communicate()
        if cmd.returncode == 0:
            try:
                tops = json.loads(stdout.decode('utf-8'))
                tops_py_version_mapping['2'] = tops
            except ValueError:
                pass

    tfp = tarfile.open(thintar, 'w:gz', dereference=True)
    try:  # cwd may not exist if it was removed but salt was run from it
        start_dir = os.getcwd()
    except OSError:
        start_dir = None
    tempdir = None
    for py_ver, tops in _six.iteritems(tops_py_version_mapping):
        for top in tops:
            base = os.path.basename(top)
            top_dirname = os.path.dirname(top)
            if os.path.isdir(top_dirname):
                os.chdir(top_dirname)
            else:
                # This is likely a compressed python .egg
                tempdir = tempfile.mkdtemp()
                egg = zipfile.ZipFile(top_dirname)
                egg.extractall(tempdir)
                top = os.path.join(tempdir, base)
                os.chdir(tempdir)
            if not os.path.isdir(top):
                # top is a single file module
                if os.path.exists(os.path.join(top_dirname, base)):
                    tfp.add(base, arcname=os.path.join('py{0}'.format(py_ver), base))
                continue
            for root, dirs, files in os.walk(base, followlinks=True):
                for name in files:
                    if not name.endswith(('.pyc', '.pyo')):
                        tfp.add(os.path.join(root, name),
                                arcname=os.path.join('py{0}'.format(py_ver), root, name))
            if tempdir is not None:
                shutil.rmtree(tempdir)
                tempdir = None
    os.chdir(thindir)
    tfp.add('salt-call')
    with salt.utils.fopen(thinver, 'w+') as fp_:
        fp_.write(salt.version.__version__)
    with salt.utils.fopen(pythinver, 'w+') as fp_:
        fp_.write(str(sys.version_info[0]))
    os.chdir(os.path.dirname(thinver))
    tfp.add('version')
    tfp.add('.thin-gen-py-version')
    if start_dir:
        os.chdir(start_dir)
    tfp.close()
    return thintar


def thin_sum(cachedir, form='sha1'):
    '''
    Return the checksum of the current thin tarball
    '''
    thintar = gen_thin(cachedir)
    return salt.utils.get_hash(thintar, form)


def gen_min(cachedir, extra_mods='', overwrite=False, so_mods='',
            python2_bin='python2', python3_bin='python3'):
    '''
    Generate the salt-min tarball and print the location of the tarball
    Optional additional mods to include (e.g. mako) can be supplied as a comma
    delimited string.  Permits forcing an overwrite of the output file as well.

    CLI Example:

    .. code-block:: bash

        salt-run min.generate
        salt-run min.generate mako
        salt-run min.generate mako,wempy 1
        salt-run min.generate overwrite=1
    '''
    mindir = os.path.join(cachedir, 'min')
    if not os.path.isdir(mindir):
        os.makedirs(mindir)
    mintar = os.path.join(mindir, 'min.tgz')
    minver = os.path.join(mindir, 'version')
    pyminver = os.path.join(mindir, '.min-gen-py-version')
    salt_call = os.path.join(mindir, 'salt-call')
    with salt.utils.fopen(salt_call, 'w+') as fp_:
        fp_.write(SALTCALL)
    if os.path.isfile(mintar):
        if not overwrite:
            if os.path.isfile(minver):
                with salt.utils.fopen(minver) as fh_:
                    overwrite = fh_.read() != salt.version.__version__
                if overwrite is False and os.path.isfile(pyminver):
                    with salt.utils.fopen(pyminver) as fh_:
                        overwrite = fh_.read() != str(sys.version_info[0])
            else:
                overwrite = True

        if overwrite:
            try:
                os.remove(mintar)
            except OSError:
                pass
        else:
            return mintar
    if _six.PY3:
        # Let's check for the minimum python 2 version requirement, 2.6
        py_shell_cmd = (
            python2_bin + ' -c \'from __future__ import print_function; import sys; '
            'print("{0}.{1}".format(*(sys.version_info[:2])));\''
        )
        cmd = subprocess.Popen(py_shell_cmd, stdout=subprocess.PIPE, shell=True)
        stdout, _ = cmd.communicate()
        if cmd.returncode == 0:
            py2_version = tuple(int(n) for n in stdout.decode('utf-8').strip().split('.'))
            if py2_version < (2, 6):
                # Bail!
                raise salt.exceptions.SaltSystemExit(
                    'The minimum required python version to run salt-ssh is "2.6".'
                    'The version reported by "{0}" is "{1}". Please try "salt-ssh '
                    '--python2-bin=<path-to-python-2.6-binary-or-higher>".'.format(python2_bin,
                                                                                stdout.strip())
                )
    elif sys.version_info < (2, 6):
        # Bail! Though, how did we reached this far in the first place.
        raise salt.exceptions.SaltSystemExit(
            'The minimum required python version to run salt-ssh is "2.6".'
        )

    tops_py_version_mapping = {}
    tops = get_tops(extra_mods=extra_mods, so_mods=so_mods)
    if _six.PY2:
        tops_py_version_mapping['2'] = tops
    else:
        tops_py_version_mapping['3'] = tops

    # TODO: Consider putting known py2 and py3 compatible libs in it's own sharable directory.
    #       This would reduce the min size.
    if _six.PY2 and sys.version_info[0] == 2:
        # Get python 3 tops
        py_shell_cmd = (
            python3_bin + ' -c \'import sys; import json; import salt.utils.thin; '
            'print(json.dumps(salt.utils.thin.get_tops(**(json.loads(sys.argv[1]))))); exit(0);\' '
            '\'{0}\''.format(json.dumps({'extra_mods': extra_mods, 'so_mods': so_mods}))
        )
        cmd = subprocess.Popen(py_shell_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        stdout, stderr = cmd.communicate()
        if cmd.returncode == 0:
            try:
                tops = json.loads(stdout)
                tops_py_version_mapping['3'] = tops
            except ValueError:
                pass
    if _six.PY3 and sys.version_info[0] == 3:
        # Get python 2 tops
        py_shell_cmd = (
            python2_bin + ' -c \'from __future__ import print_function; '
            'import sys; import json; import salt.utils.thin; '
            'print(json.dumps(salt.utils.thin.get_tops(**(json.loads(sys.argv[1]))))); exit(0);\' '
            '\'{0}\''.format(json.dumps({'extra_mods': extra_mods, 'so_mods': so_mods}))
        )
        cmd = subprocess.Popen(py_shell_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        stdout, stderr = cmd.communicate()
        if cmd.returncode == 0:
            try:
                tops = json.loads(stdout.decode('utf-8'))
                tops_py_version_mapping['2'] = tops
            except ValueError:
                pass

    tfp = tarfile.open(mintar, 'w:gz', dereference=True)
    try:  # cwd may not exist if it was removed but salt was run from it
        start_dir = os.getcwd()
    except OSError:
        start_dir = None
    tempdir = None

    # This is the absolute minimum set of files required to run salt-call
    min_files = (
        'salt/__init__.py',
        'salt/utils',
        'salt/utils/__init__.py',
        'salt/utils/validate',
        'salt/utils/validate/__init__.py',
        'salt/utils/validate/path.py',
        'salt/utils/decorators',
        'salt/utils/decorators/__init__.py',
        'salt/utils/cache.py',
        'salt/utils/xdg.py',
        'salt/utils/odict.py',
        'salt/utils/minions.py',
        'salt/utils/dicttrim.py',
        'salt/utils/sdb.py',
        'salt/utils/migrations.py',
        'salt/utils/files.py',
        'salt/utils/parsers.py',
        'salt/utils/locales.py',
        'salt/utils/lazy.py',
        'salt/utils/s3.py',
        'salt/utils/dictupdate.py',
        'salt/utils/verify.py',
        'salt/utils/args.py',
        'salt/utils/kinds.py',
        'salt/utils/xmlutil.py',
        'salt/utils/debug.py',
        'salt/utils/jid.py',
        'salt/utils/openstack',
        'salt/utils/openstack/__init__.py',
        'salt/utils/openstack/swift.py',
        'salt/utils/async.py',
        'salt/utils/process.py',
        'salt/utils/jinja.py',
        'salt/utils/rsax931.py',
        'salt/utils/context.py',
        'salt/utils/minion.py',
        'salt/utils/error.py',
        'salt/utils/aws.py',
        'salt/utils/timed_subprocess.py',
        'salt/utils/zeromq.py',
        'salt/utils/schedule.py',
        'salt/utils/url.py',
        'salt/utils/yamlencoding.py',
        'salt/utils/network.py',
        'salt/utils/http.py',
        'salt/utils/gzip_util.py',
        'salt/utils/vt.py',
        'salt/utils/templates.py',
        'salt/utils/aggregation.py',
        'salt/utils/yamlloader.py',
        'salt/utils/event.py',
        'salt/serializers',
        'salt/serializers/__init__.py',
        'salt/serializers/yamlex.py',
        'salt/template.py',
        'salt/_compat.py',
        'salt/loader.py',
        'salt/client',
        'salt/client/__init__.py',
        'salt/ext',
        'salt/ext/__init__.py',
        'salt/ext/six.py',
        'salt/ext/ipaddress.py',
        'salt/version.py',
        'salt/syspaths.py',
        'salt/defaults',
        'salt/defaults/__init__.py',
        'salt/defaults/exitcodes.py',
        'salt/renderers',
        'salt/renderers/__init__.py',
        'salt/renderers/jinja.py',
        'salt/renderers/yaml.py',
        'salt/modules',
        'salt/modules/__init__.py',
        'salt/modules/test.py',
        'salt/modules/selinux.py',
        'salt/modules/cmdmod.py',
        'salt/minion.py',
        'salt/pillar',
        'salt/pillar/__init__.py',
        'salt/textformat.py',
        'salt/log',
        'salt/log/__init__.py',
        'salt/log/handlers',
        'salt/log/handlers/__init__.py',
        'salt/log/mixins.py',
        'salt/log/setup.py',
        'salt/cli',
        'salt/cli/__init__.py',
        'salt/cli/caller.py',
        'salt/cli/daemons.py',
        'salt/cli/salt.py',
        'salt/cli/call.py',
        'salt/fileserver',
        'salt/fileserver/__init__.py',
        'salt/transport',
        'salt/transport/__init__.py',
        'salt/transport/client.py',
        'salt/exceptions.py',
        'salt/grains',
        'salt/grains/__init__.py',
        'salt/grains/extra.py',
        'salt/scripts.py',
        'salt/state.py',
        'salt/fileclient.py',
        'salt/crypt.py',
        'salt/config.py',
        'salt/beacons',
        'salt/beacons/__init__.py',
        'salt/payload.py',
        'salt/output',
        'salt/output/__init__.py',
        'salt/output/nested.py',
    )

    for py_ver, tops in _six.iteritems(tops_py_version_mapping):
        for top in tops:
            base = os.path.basename(top)
            top_dirname = os.path.dirname(top)
            if os.path.isdir(top_dirname):
                os.chdir(top_dirname)
            else:
                # This is likely a compressed python .egg
                tempdir = tempfile.mkdtemp()
                egg = zipfile.ZipFile(top_dirname)
                egg.extractall(tempdir)
                top = os.path.join(tempdir, base)
                os.chdir(tempdir)
            if not os.path.isdir(top):
                # top is a single file module
                tfp.add(base, arcname=os.path.join('py{0}'.format(py_ver), base))
                continue
            for root, dirs, files in os.walk(base, followlinks=True):
                for name in files:
                    if name.endswith(('.pyc', '.pyo')):
                        continue
                    if root.startswith('salt') and os.path.join(root, name) not in min_files:
                        continue
                    tfp.add(os.path.join(root, name),
                            arcname=os.path.join('py{0}'.format(py_ver), root, name))
            if tempdir is not None:
                shutil.rmtree(tempdir)
                tempdir = None

    os.chdir(mindir)
    tfp.add('salt-call')
    with salt.utils.fopen(minver, 'w+') as fp_:
        fp_.write(salt.version.__version__)
    with salt.utils.fopen(pyminver, 'w+') as fp_:
        fp_.write(str(sys.version_info[0]))
    os.chdir(os.path.dirname(minver))
    tfp.add('version')
    tfp.add('.min-gen-py-version')
    if start_dir:
        os.chdir(start_dir)
    tfp.close()
    return mintar


def min_sum(cachedir, form='sha1'):
    '''
    Return the checksum of the current thin tarball
    '''
    mintar = gen_min(cachedir)
    return salt.utils.get_hash(mintar, form)



# -*- coding: utf-8 -*-
'''Test cases for the ``ldap`` state module

This code is gross.  I started out trying to remove some of the
duplicate code in the test cases, and before I knew it the test code
was an ugly second implementation.

I'm leaving it for now, but this should really be gutted and replaced
with something sensible.
'''

from __future__ import absolute_import

import copy
import salt.ext.six as six
import salt.states.ldap

from salttesting import skipIf, TestCase
from salttesting.helpers import ensure_in_syspath
from salttesting.mock import (
    NO_MOCK,
    NO_MOCK_REASON,
    patch,
)

ensure_in_syspath('../../')


# emulates the LDAP database.  each key is the DN of an entry and it
# maps to a dict which maps attribute names to sets of values.
db = {}


def _init_db(newdb=None):
    if newdb is None:
        newdb = {}
    global db
    db = newdb


def _complex_db():
    return {
        'dnfoo': {
            'attrfoo1': set((
                'valfoo1.1',
                'valfoo1.2',
            )),
            'attrfoo2': set((
                'valfoo2.1',
            )),
        },
        'dnbar': {
            'attrbar1': set((
                'valbar1.1',
                'valbar1.2',
            )),
            'attrbar2': set((
                'valbar2.1',
            )),
        },
    }


class _dummy_ctx(object):
    def __init__(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        pass


def _dummy_connect(connect_spec):
    return _dummy_ctx()


def _dummy_search(connect_spec, base, scope):
    if base not in db:
        return {}
    return {base: dict(((attr, sorted(db[base][attr]))
                        for attr in db[base]
                        if len(db[base][attr])))}


def _dummy_add(connect_spec, dn, attributes):
    assert dn not in db
    assert len(attributes)
    db[dn] = {}
    for attr, vals in six.iteritems(attributes):
        assert len(vals)
        db[dn][attr] = set(vals)
    return True


def _dummy_delete(connect_spec, dn):
    assert dn in db
    del db[dn]
    return True


def _dummy_change(connect_spec, dn, before, after):
    assert before != after
    assert len(before)
    assert len(after)
    assert dn in db
    e = db[dn]
    assert e == before
    all_attrs = set()
    all_attrs.update(before)
    all_attrs.update(after)
    directives = []
    for attr in all_attrs:
        if attr not in before:
            assert attr in after
            assert len(after[attr])
            directives.append(('add', attr, after[attr]))
        elif attr not in after:
            assert attr in before
            assert len(before[attr])
            directives.append(('delete', attr, ()))
        else:
            assert len(before[attr])
            assert len(after[attr])
            to_del = before[attr] - after[attr]
            if len(to_del):
                directives.append(('delete', attr, to_del))
            to_add = after[attr] - before[attr]
            if len(to_add):
                directives.append(('add', attr, to_add))
    return _dummy_modify(connect_spec, dn, directives)


def _dummy_modify(connect_spec, dn, directives):
    assert dn in db
    e = db[dn]
    for op, attr, vals in directives:
        if op == 'add':
            assert len(vals)
            existing_vals = e.setdefault(attr, set())
            for val in vals:
                assert val not in existing_vals
                existing_vals.add(val)
        elif op == 'delete':
            assert attr in e
            existing_vals = e[attr]
            assert len(existing_vals)
            if not len(vals):
                del e[attr]
                continue
            for val in vals:
                assert val in existing_vals
                existing_vals.remove(val)
            if not len(existing_vals):
                del e[attr]
        elif op == 'replace':
            e.pop(attr, None)
            e[attr] = set(vals)
        else:
            raise ValueError()
    return True


def _dump_db(d=None):
    if d is None:
        d = db
    return dict(((dn, dict(((attr, sorted(d[dn][attr]))
                            for attr in d[dn])))
                 for dn in d))


@skipIf(NO_MOCK, NO_MOCK_REASON)
class LDAPTestCase(TestCase):

    def setUp(self):
        __opts = getattr(salt.states.ldap, '__opts__', {})
        salt.states.ldap.__opts__ = __opts
        __salt = getattr(salt.states.ldap, '__salt__', {})
        salt.states.ldap.__salt__ = __salt
        self.patchers = [
            patch.dict('salt.states.ldap.__opts__', {'test': False}),
        ]
        for f in ('connect', 'search', 'add', 'delete', 'change', 'modify'):
            self.patchers.append(
                patch.dict('salt.states.ldap.__salt__',
                           {'ldap3.' + f: globals()['_dummy_' + f]}))
        for p in self.patchers:
            p.start()
        self.maxDiff = None

    def tearDown(self):
        for p in reversed(self.patchers):
            p.stop()

    def _test_helper(self, init_db, expected_ret, replace,
                     delete_others=False):
        _init_db(copy.deepcopy(init_db))
        old = _dump_db()
        new = _dump_db()
        expected_db = copy.deepcopy(init_db)
        for dn, attrs in six.iteritems(replace):
            for attr, vals in six.iteritems(attrs):
                if len(vals):
                    new.setdefault(dn, {})[attr] = sorted(set(vals))
                    expected_db.setdefault(dn, {})[attr] = set(vals)
                elif dn in expected_db:
                    new[dn].pop(attr, None)
                    expected_db[dn].pop(attr, None)
            if not len(expected_db.get(dn, {})):
                new.pop(dn, None)
                expected_db.pop(dn, None)
        if delete_others:
            dn_to_delete = set()
            for dn, attrs in six.iteritems(expected_db):
                if dn in replace:
                    to_delete = set()
                    for attr, vals in six.iteritems(attrs):
                        if attr not in replace[dn]:
                            to_delete.add(attr)
                    for attr in to_delete:
                        del attrs[attr]
                        del new[dn][attr]
                    if not len(attrs):
                        dn_to_delete.add(dn)
            for dn in dn_to_delete:
                del new[dn]
                del expected_db[dn]
        name = 'ldapi:///'
        expected_ret['name'] = name
        expected_ret.setdefault('result', True)
        expected_ret.setdefault('comment', 'Successfully updated LDAP entries')
        expected_ret.setdefault('changes', dict(
            ((dn, {'old': dict((attr, vals)
                               for attr, vals in six.iteritems(old[dn])
                               if vals != new.get(dn, {}).get(attr, ()))
                          if dn in old else None,
                   'new': dict((attr, vals)
                               for attr, vals in six.iteritems(new[dn])
                               if vals != old.get(dn, {}).get(attr, ()))
                          if dn in new else None})
             for dn in replace
             if old.get(dn, {}) != new.get(dn, {}))))
        entries = [{dn: [{'replace': attrs},
                         {'delete_others': delete_others}]}
                   for dn, attrs in six.iteritems(replace)]
        actual = salt.states.ldap.managed(name, entries)
        self.assertDictEqual(expected_ret, actual)
        self.assertDictEqual(expected_db, db)

    def _test_helper_success(self, init_db, replace, delete_others=False):
        self._test_helper(init_db, {}, replace, delete_others)

    def _test_helper_nochange(self, init_db, replace, delete_others=False):
        expected = {
            'changes': {},
            'comment': 'LDAP entries already set',
        }
        self._test_helper(init_db, expected, replace, delete_others)

    def test_managed_empty(self):
        _init_db()
        name = 'ldapi:///'
        expected = {
            'name': name,
            'changes': {},
            'result': True,
            'comment': 'LDAP entries already set',
        }
        actual = salt.states.ldap.managed(name, {})
        self.assertDictEqual(expected, actual)

    def test_managed_add_entry(self):
        self._test_helper_success(
            {},
            {'dummydn': {'foo': ['bar', 'baz']}})

    def test_managed_add_attr(self):
        self._test_helper_success(
            _complex_db(),
            {'dnfoo': {'attrfoo3': ['valfoo3.1']}})

    def test_managed_simplereplace(self):
        self._test_helper_success(
            _complex_db(),
            {'dnfoo': {'attrfoo1': ['valfoo1.3']}})

    def test_managed_deleteattr(self):
        self._test_helper_success(
            _complex_db(),
            {'dnfoo': {'attrfoo1': []}})

    def test_managed_deletenonexistattr(self):
        self._test_helper_nochange(
            _complex_db(),
            {'dnfoo': {'dummyattr': []}})

    def test_managed_deleteentry(self):
        self._test_helper_success(
            _complex_db(),
            {'dnfoo': {}},
            True)

    def test_managed_deletenonexistentry(self):
        self._test_helper_nochange(
            _complex_db(),
            {'dummydn': {}},
            True)

    def test_managed_deletenonexistattrinnonexistentry(self):
        self._test_helper_nochange(
            _complex_db(),
            {'dummydn': {'dummyattr': []}})

    def test_managed_add_attr_delete_others(self):
        self._test_helper_success(
            _complex_db(),
            {'dnfoo': {'dummyattr': ['dummyval']}},
            True)

    def test_managed_no_net_change(self):
        self._test_helper_nochange(
            _complex_db(),
            {'dnfoo': {'attrfoo1': ['valfoo1.2', 'valfoo1.1']}})

    def test_managed_repeated_values(self):
        self._test_helper_success(
            {},
            {'dummydn': {'dummyattr': ['dummyval', 'dummyval']}})


if __name__ == '__main__':
    from integration import run_tests
    run_tests(LDAPTestCase, needs_daemon=False)
